#!python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import os
import sys
import re

import cvslib
import uriparse
import process
import timeline
import koprocessutils

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject
from xpcom.server import WrapObject, UnwrapObject

class KoCVS:
    _com_interfaces_ = [components.interfaces.koICVS, components.interfaces.nsIObserver]
    _reg_desc_ = "Komodo CVS Support"
    _reg_contractid_ = "@activestate.com/koCVS;1"
    _reg_clsid_ = "{CE39F3D2-7103-493f-8A77-EBD15071CD57}"
    
    def __init__(self):
        # generic cvs instance to do work with
        timeline.enter('KoCVS.__init__')
        self.cvs = cvslib.CVS()
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)
        self.fileStatusSvc = components.classes["@activestate.com/koFileService;1"].\
                getService(components.interfaces.koIFileService);
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
        self._encodingSvc = components.classes['@activestate.com/koEncodingServices;1'].\
                                getService(components.interfaces.koIEncodingServices)
        self.reasonNotFunctional = None
        self.isFunctional = None
        self._reVersion = re.compile(r'System\s\((.*?)\)\s([\d|\.]*?)\s',re.M|re.VERBOSE)
        self._globalPrefs.prefObserverService.addObserver(WrapObject(self,components.interfaces.nsIObserver),'cvsExecutable',1)

        self._userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)
        self._executable = None
        self._setExecutableFromPrefs()
        self.redetermineIfFunctional()
        timeline.leave('KoCVS.__init__')

    def observe(self, subject, topic, value):
        if topic == 'cvsExecutable':
            self._setExecutableFromPrefs()
    
    def _setExecutableFromPrefs(self):
        executable = self._globalPrefs.getStringPref('cvsExecutable')
        if self._executable != executable:
            self.set_executable(executable)
            try:
                self._observerSvc.notifyObservers(self, 'subsystem_changed', 'cvs')
            except COMException:
                pass # Ignore error if there is no observer.
        
    def set_executable(self, executable):
        self._executable = executable
        # we want to allow self._executable to be None, but
        # we cannot allow self.cvs._cvs to be None
        if self._executable:
            self.cvs._cvs = self._executable
        else:
            self.cvs._cvs = 'cvs'
        
    def get_executable(self):
        return self._executable
    
    def redetermineIfFunctional(self):
        # XXX if there are multiple cvs executables, we should determine
        # which ones are usable, which is the latest version, etc.,
        # and stuff that information someplace
        timeline.enter('redetermineIfFunctional')
        import which
        assert which._version_ >= (1,0,0)
        cvsInfoEx = components.classes["@activestate.com/koCVSInfoEx;1"].\
                    createInstance(components.interfaces.koIAppInfoEx)
        if not self._executable:
            try:
                if sys.platform.startswith('win'):
                    exts = ['.exe']
                else:
                    exts = None
                cvs = which.which('cvs', exts=exts, path=self._userPath)
            except which.WhichError:
                pass
            else:
                # this call sets cvsInfoEx.executablePath
                cvsInfoEx.getInstallationPathFromBinary(cvs)
        else:
            cvsInfoEx.executablePath = self._executable;

        if not cvsInfoEx.executablePath:
            self.isFunctional = 0
            self.reasonNotFunctional = "no 'cvs' executable could be found"
        else:
            version = cvsInfoEx.version
            self.isFunctional = 0
            if not version:
                self.reasonNotFunctional = \
                    "the executable is not a CVS executable or "\
                    "is otherwise incompatible with Komodo"
            else:
                self.isFunctional = 1
                self.reasonNotFunctional = None
        # XXX it seems to me that determining whether an executable
        # is functional does not mean we should change the subsystem
        try:
            self._observerSvc.notifyObservers(self, 'subsystem_changed', 'cvs')
        except COMException:
            pass # Ignore error if there is no observer.
        timeline.leave('redetermineIfFunctional')

    def _splitFile(self, file,splitdir):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        try:
            path = uriparse.URIToLocalPath(file);
            if path:
                if not splitdir and os.path.isdir(path):
                    return path,''
                path = os.path.normpath(path)
                basedir = os.path.dirname(path)
                filename = os.path.basename(path)
                #print "_splitfile %s:%s:%s"% (path,basedir,filename)
                # make sure the path is in cvs
                root = self.cvs.getFileRoot(basedir)
                if root:
                    return basedir, filename
        except cvslib.CVSLibError, e:
            # we'll ignore files that are not in cvs
            pass
        return None,None
    
    def _splitFiles(self, files, splitdir=0):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        _files = {}
        for file in files:
            basedir,filename = self._splitFile(file,splitdir)
            if basedir:
                if not basedir in _files:
                    _files[basedir] = []
                _files[basedir].append(filename)
        return _files

    # Regex used to update the "Index: xyz..." to have a full path reference
    diffUpdatePathRegex = re.compile("^Index: ", re.MULTILINE)

#    cvs diff [-lNR] [rcsdiff-options]
#    [[-r rev1 | -D date1] [-r rev2 | -D date2]] [files...] 
#	-l	Local directory only, not recursive
#	-R	Process directories recursively.
#	-D d1	Diff revision for date against working file.
#	-D d2	Diff rev1/date1 against date2.
#	-N	include diffs for added and removed files.
#	-r rev1	Diff revision for rev1 against working file.
#	-r rev2	Diff rev1/date1 against rev2.
#	--ifdef=arg	Output diffs in ifdef format.
    def diff(self, files, options, external):
        #print "cvs diff ",repr(files)
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        diff = ''
        for basedir in _files.keys():
            #print "diff files [%s] :%s" % (basedir,repr(_files[basedir]))
            try:
                result = self.cvs.diff(files=_files[basedir],
                                   cwd=basedir,
                                   diffOpts=options.split())
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            
            if 'stdout' in result:
                replaceStr = "Index: %s%s" % (os.path.abspath(basedir), os.sep)
                replaceStr = replaceStr.replace("\\", "\\\\")
                diff += self.diffUpdatePathRegex.sub(replaceStr,
                                                     result['stdout'])
            elif 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))

        # Convert line-endings to the currently expected
        import eollib
        eolpref = self._globalPrefs.getStringPref('endOfLine')
        try:
            desiredEOL = eollib.eolPref2eol[eolpref]
        except KeyError:
            desiredEOL = eollib.EOL_PLATFORM
        diff = eollib.convertToEOLFormat(diff, desiredEOL)

        # Try encoding it with komodo's unicode encoding service
        try:
            diff, encoding, bom = self._encodingSvc.getUnicodeEncodedString(diff)
        except Exception, e:
            # No go, we could not work out the encoding
            self.lastErrorSvc.setLastError(0, str(e))
            return None

        return diff

    # Diff two revisions of the given file
    def diffRevisions(self, fileuri1, rev1, fileuri2, rev2, localfilepath,
                      options, external):
        #print "cvs diffRevisions ", repr(fileuri1), rev1, repr(fileuri2), rev2
        revOptions = []
        for revision in (rev1, rev2):
            if revision:
                revOptions.append("-r %s" %(revision))
        revOptions = " ".join(revOptions)
        if options:
            options += " %s" % (revOptions)
        else:
            options = revOptions
        return self.diff([fileuri1], options, external)

#    Usage: cvs log [-lRhtNb] [-r[revisions]] [-d dates] [-s states]
#        [-w[logins]] [files...]
#            -l      Local directory only, no recursion.
#            -R      Only print name of RCS file.
#            -h      Only print header.
#            -t      Only print header and descriptive text.
#            -N      Do not list tags.
#            -S      Do not print name/header if no revisions selected.
#            -b      Only list revisions on the default branch.
#            -r[revisions]   A comma-separated list of revisions to print:
#               rev1:rev2   Between rev1 and rev2, including rev1 and rev2.
#               rev1::rev2  Between rev1 and rev2, excluding rev1.
#               rev:        rev and following revisions on the same branch.
#               rev::       After rev on the same branch.
#               :rev        rev and previous revisions on the same branch.
#               ::rev       rev and previous revisions on the same branch.
#               rev         Just rev.
#               branch      All revisions on the branch.
#               branch.     The last revision on the branch.
#            -d dates        A semicolon-separated list of dates
#                            (D1<D2 for range, D for latest before).
#            -s states       Only list revisions with specified states.
#            -w[logins]      Only list revisions checked in by specified logins.
    # Example output:
    #
    #C:\src\komodo\tests\scc\CVS\pyDes>cvs log README.txt
    #
    #RCS file: /cvsroot/pydes/pyDes/README.txt,v
    #Working file: README.txt
    #head: 1.2
    #branch:
    #locks: strict
    #access list:
    #symbolic names:
    #keyword substitution: kv
    #total revisions: 2;     selected revisions: 2
    #description:
    #----------------------------
    #revision 1.2
    #date: 2005/09/13 17:25:46;  author: twhiteman;  state: Exp;  lines: +3 -3
    #Version 1.2 - Fix Triple DES CCB mode.
    #----------------------------
    #revision 1.1
    #date: 2005/08/29 17:37:01;  author: twhiteman;  state: Exp;
    #Initial creation.
    #=============================================================================

    #re_log_head = re.compile(
    #        r'''RCS file:\s*(?P<repository_file>.*?)[\r\n]*?'''\
    #         '''Working file:\s*(?P<local_file>.*?)[\r\n]*?'''\
    #         '''head:\s*(?P<head>.*?)[\r\n]*?'''\
    #         '''branch:\s*(?P<branch>.*?)[\r\n]*?'''\
    #         '''locks:\s*(?P<locks>.*?)[\r\n]*?'''\
    #         '''access list:\s*(?P<access_list>.*?)[\r\n]*?'''\
    #         '''symbolic names:\s*(?P<symbolic_names>.*?)[\r\n]*?'''\
    #         '''keyword substitution:\s*(?P<keyword_substitution>.*?)[\r\n]*?'''\
    #         '''total revisions: \s*(?P<total_revisions>.*?);\s*selected revisions:\s*(?P<selected_revisions>.*?)[\r\n]*?'''\
    #         '''description:\s*[\r\n]*?'''
    #)
    re_log_revisions    = re.compile(
        r'''----------------------------\s*[\r\n]*'''\
        '''revision\s+(?P<revision>.*?)[\r\n]*'''\
        '''date: \s*(?P<date>.*?);\s*author:\s*(?P<author>.*?);\s*state:\s*(?P<state>.*?).*[\r\n]*'''
    )

    def history(self, fileuri, options):
        #print "cvs log ", options, repr(fileuri)
        self.lastErrorSvc.setLastError(0, '')

        basedir, filename = self._splitFile(fileuri, 1)
        try:
            result = self.cvs.log(files=[filename], cwd=basedir)
        except cvslib.CVSLibError, e:
            self.lastErrorSvc.setLastError(0, str(e))
            return ''

        rawerror = result.get('stderr', "")
        if rawerror and rawerror.find("Empty password used") == -1:
            # There was some type of error
            self.lastErrorSvc.setLastError(0, str(rawerror))
            #print "stderr:", rawerror
            return None

        output = result['stdout']

        #headMatch = self.re_log_head.search(output)
        #if not headMatch:
        #    rawerror = "Could not parse cvs log message"
        #    self.lastErrorSvc.setLastError(0, rawerror)
        #    #print "stderr:", rawerror
        #    return None

        # Now, split up the history rows
        #
        koHistoryList = []
        #revisionsText = output[headMatch.end():]
        revisionsText = output
        matchIterator = self.re_log_revisions.finditer(revisionsText)
        try:
            match = matchIterator.next()
        except StopIteration:
            match = None
        while match:
            koHistory = components.classes["@activestate.com/koSCCHistoryItem;1"].\
                            createInstance(components.interfaces.koISCCHistoryItem);
            koHistory.version = match.group('revision')
            koHistory.author  = match.group('author')
            koHistory.date    = match.group('date')
            koHistory.action  = match.group('state')
            koHistory.uri     = fileuri
            text_start = match.end()
            try:
                match = matchIterator.next()
            except StopIteration:
                match = None
            message = ''
            if match:
                message = revisionsText[text_start:match.start()]
            else:
                leftOverText = revisionsText[text_start:]
                endMarker = leftOverText.rfind("=============================================================================")
                if endMarker > 0:
                    leftOverText = leftOverText[:endMarker]
                message = leftOverText
            # Set the history message
            try:
                koHistory.message = message
            except UnicodeDecodeError:
                # Try encoding it with komodo's unicode encoding service
                try:
                    buffer, encoding, bom = self._encodingSvc.getUnicodeEncodedString(message)
                    koHistory.message = buffer
                except Exception, e:
                    # No go, show that there is something wrong then
                    koHistory.message = "Komodo: Encoding error: Could not interpret the history message"
            koHistoryList.append(koHistory)
        return koHistoryList

#    cvs add [-k rcs-kflag] [-m message] files...
#	-k	Use "rcs-kflag" to add the file with the specified kflag.
#	-m	Use "message" for the creation log.
    def add(self, files, mode, message):
        #add(self, files, mode=None, msg=None, cwd=None)
        # first we figure out what repositories files are in
        # and split them up into seperate lists
        # we init the cvs object with the root we retreived
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files,1)
        output = ''
        for basedir in _files.keys():
            try:
                result = self.cvs.add(files=_files[basedir], mode=mode, msg=message, cwd=basedir)
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            
            if 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))
                output += result['stderr']
            if 'stdout' in result:
                output += result['stdout']
            
        # do notification that status may have changed
        # we have to use the original urls passed in since
        # they are modified to local paths in _splitFiles
        for file in files:
            if file: self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
    
#    cvs commit [-nRlf] [-m msg | -F logfile] [-r rev] files...
#	-n	Do not run the module program (if any).
#	-R	Process directories recursively.
#	-l	Local directory only (not recursive).
#	-f	Force the file to be committed; disables recursion.
#	-F file	Read the log message from file.
#	-m msg	Log message.
#	-r rev	Commit to this branch or trunk revision.
    def commit(self, files, message, options):
        #print "cvs commit ",repr(files)
        # save the message to a temp file
        msgfile = self.fileStatusSvc.makeTempFile('.cvs','w')
        msgfile.puts(message)
        msgfile.flush()
        msgfile.close()
        msgpath = msgfile.path

        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            #print "commit files [%s] :%s" % (basedir,repr(_files[basedir]))
            try:
                result = self.cvs.commit(_files[basedir],msgFile=msgpath,cwd=basedir)
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            #print repr(result)
            if 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))
                output += result['stderr']
            if 'stdout' in result:
                output += result['stdout']
        
        del msgfile
        
        for file in files:
            if file: self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
    
#    cvs remove [-flR] [files...]
#	-f	Delete the file before removing it.
#	-l	Process this directory only (not recursive).
#	-R	Process directories recursively.
    def remove(self, files, force, recursive):
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            try:
                result = self.cvs.remove(_files[basedir], force, recursive,cwd=basedir)
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))
                output += result['stderr']
            if 'stdout' in result:
                output += result['stdout']
            
        for file in files:
            if file: self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
    
#    cvs update [-APdflRp] [-k kopt] [-r rev|-D date] [-j rev]
#    [-I ign] [-W spec] [files...]
#	-A	Reset any sticky tags/date/kopts.
#	-P	Prune empty directories.
#	-d	Build directories, like checkout does.
#	-f	Force a head revision match if tag/date not found.
#	-l	Local directory only, no recursion.
#	-R	Process directories recursively.
#	-p	Send updates to standard output (avoids stickiness).
#	-k kopt	Use RCS kopt -k option on checkout.
#	-r rev	Update using specified revision/tag (is sticky).
#	-D date	Set date to update from (is sticky).
#	-j rev	Merge in changes made between current revision and rev.
#	-I ign	More files to ignore (! to reset).
#	-W spec	Wrappers specification line.
    def update(self, files, options):
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            try:
                result = self.cvs.update(_files[basedir],cwd=basedir)
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))
                
            # build a better status output for komodo
            if 'files' in result:
                for file in result['files']:
                    localFileName = os.path.normpath(os.path.join(basedir,file['file']))
                    url = uriparse.localPathToURI(localFileName)
                    #print "%s %s %s" %(file['status'],file['file'],url)
                    output += "%s %s %s\n" % (file['status'],file['file'],url)

        for file in files:
            if file: self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output

    def revert(self, files, options):
        # XXX unfortunately this function requires knowledge of
        # item properties
        self.lastErrorSvc.setLastError(0, '')
        output = ''
        for file in files:
            kofile = self.fileStatusSvc.getFileFromURI(file)
            
            # splitfile does more than initial appearances
            basedir,filename = self._splitFile(file,0)
            #print "attempting to revert file ",file
            if kofile and kofile.sccAction:
                # we are edit, delete or sync
                try:
                    if kofile.sccAction == 'edit':
                        result = self.cvs.unedit(files=filename,cwd=basedir)
                        result = self.cvs.update(files=filename,clean=1,cwd=basedir)
                    elif kofile.sccAction == 'delete':
                        #print "adding deleted file back into cvs"
                        result = self.cvs.add(files=filename,cwd=basedir)
                    else:
                        #print "doing CLEAN update"
                        result = self.cvs.update(files=filename,clean=1,cwd=basedir)
                except cvslib.CVSLibError, e:
                    self.lastErrorSvc.setLastError(0, str(e))
                    return ''
                if 'stderr' in result:
                    output += result['stderr']
                    self.lastErrorSvc.setLastError(0, str(result['stderr']))
                if 'stdout' in result:
                    output += result['stdout']
                self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
   
    def edit(self, files):
        # XXX unfortunately this function requires knowledge of
        # item properties
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            # ,cwd=basedir
            try:
                result = self.cvs.edit(_files[basedir], cwd=basedir)
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))
                output += result['stderr']
            if 'stdout' in result:
                output += result['stdout']

        for file in files:
            if file: self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
