#!python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import sys
import os
import string
import re, urllib
import urlparse, mimetools, mimetypes
import threading
import tempfile
import listcmd

if sys.platform.startswith("win"):
    import win32api
    # constants pulled from win32con to save memory
    VER_PLATFORM_WIN32_WINDOWS = 1
    isWin9x = win32api.GetVersionEx()[3] == VER_PLATFORM_WIN32_WINDOWS

import process
import koprocessutils
from invocationutils import *
import logging

from xpcom import components, ServerException, COMException, nsError
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject


#---- globals

log = logging.getLogger("koInvokeInterpreters")
#log.setLevel(logging.DEBUG)

INVOKE_RUN = components.interfaces.koIInvocation.INVOKE_RUN
INVOKE_DEBUG = components.interfaces.koIInvocation.INVOKE_DEBUG
INVOKE_BUILD = components.interfaces.koIInvocation.INVOKE_BUILD
INVOKE_INTERACTIVE = components.interfaces.koIInvocation.INVOKE_INTERACTIVE

invoke_names = {
    INVOKE_RUN: "Run without Debugging",
    INVOKE_DEBUG: "Debugging",
    INVOKE_BUILD: "Build/Make",
    INVOKE_INTERACTIVE: "Interactive Shell"
}

#---- support routines

def _escapeArg(arg):
    """Escape the given command line argument for the shell."""
    return arg.replace('"', r'\"')

def _joinArgv(argv):
    r"""Join an arglist to a string appropriate for running.

        >>> import os
        >>> _joinArgv(['foo', 'bar "baz'])
        'foo "bar \\"baz"'
    """
    cmdstr = ""
    specialChars = [';', ' ', '=']
    for arg in argv:
        if arg == '':
            cmdstr += '""'
        else:
            for ch in specialChars:
                if ch in arg:
                    cmdstr += '"%s"' % _escapeArg(arg)
                    break
            else:
                cmdstr += _escapeArg(arg)
        cmdstr += ' '
    if cmdstr.endswith(' '): cmdstr = cmdstr[:-1]  # strip trailing space
    return cmdstr



#---- Component implementations

class koInterpreterInvocation:
    _com_interfaces_ = components.interfaces.koIInvocation
    def __init__(self):
        self.error_no_interpreter = "No %s interpreter is available." % self._invocation_name_
        self.error_no_preferences = "Preferences for debugging %s are not initialized." % self._invocation_name_
        self.error_invocation_not_supported = self._invocation_name_ + " does not support '%s'."

        self.currentCategoryPrefs = None
        self.currentInstancePrefs = None
        self.prefService = components.classes["@activestate.com/koPrefService;1"].getService();
        self.name = self._invocation_name_
        self.process = None
        self._environUtils = components.classes[
            "@activestate.com/koEnvironUtils;1"] \
            .getService(components.interfaces.koIEnvironUtils)
        self.infoSvc = components.classes["@activestate.com/koInfoService;1"] \
            .getService(components.interfaces.koIInfoService);
        proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"]\
            .getService(components.interfaces.nsIProxyObjectManager)
        observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerSvc = getProxyForObject(None,
            components.interfaces.nsIObserverService,
            observerSvc, PROXY_ALWAYS | PROXY_SYNC)
        self.komodoVersion = self.infoSvc.version

        self.stdin = self.stdout = self.stderr = None
        self._usecgi = 0
        self._postData = None
        self.cookie = self._makeCookie()

    def needsBuild( self ):
        # Result: boolean
        # we never need a separate build process.
        return 0

    def invoke(self, invokeType):
        self._checkValid(invokeType)
        if invokeType == INVOKE_DEBUG:
            return self._launchForDebug()
        elif invokeType == INVOKE_RUN:
            return self._launchForRun()
        elif invokeType == INVOKE_INTERACTIVE:
            return self._launchForInteractive()
        else:
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE)

    def get_supportedInvokeTypes(self):
        return INVOKE_DEBUG | INVOKE_RUN | INVOKE_INTERACTIVE

    def _determineExecutable(self):
        return None

    def validInterpreter(self, invokeType):
        # Attempt to locate the executable if none specified.
        if not (invokeType & self.get_supportedInvokeTypes()):
            return 0, "", self.error_invocation_not_supported % invoke_names[invokeType]

        exe = self._determineExecutable()
        if not exe:
            return 1, "", self.error_no_interpreter
        self.currentCategoryPrefs.setStringPref("executable", exe)
        return 1, None, None
        
    def validate(self, invokeType):
        if self.currentInstancePrefs is None or self.currentCategoryPrefs is None:
            return 0, "", self.error_no_preferences

        res = self.validInterpreter(invokeType)
        if not res[0]: return res

        return self._do_validate(invokeType)

    def _checkValid(self, invokeType):
        ok, id, message = self.validate(invokeType)
        if not ok:
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, message)

    def terminate(self, exit_code, use_force):
        try:
            self.process.kill()
            return 1
        except COMException:
            pass
        return 0

    def _fillExtraDefaultPreferences(self, cat, inst):
        pass
    
    def _makeCookie(self):
        from base64 import encodestring
        import random
        
        cooklen=8
        if os.path.exists("/dev/random"):
            rand=open("/dev/random", "r")
            data=rand.read(cooklen)
        else:
            data=""
            for i in range(cooklen):
                data = data + chr(random.randint(0,127))
        return string.strip(encodestring(data))

    def getDefaultPreferences( self ):
        cat = components.classes["@activestate.com/koPreferenceSet;1"].createInstance()
        cat.setStringPref("executable", "") # validate fills this correctly if not done before.

        # we have to get the dynamic port from the listener
        manager = components.classes["@activestate.com/koDBGPManager;1"].\
           getService(components.interfaces.koIDBGPManager);
        port = manager.port

        cat.setLongPref("debugger.listener-port", port)

        inst = components.classes["@activestate.com/koPreferenceSet;1"].createInstance()
        inst.setStringPref("executable-params", "") # "Interpreter Arguments" in UI
        inst.setBooleanPref("use-console", 0) # "Debug in separate console"
        inst.setStringPref("filename", "")    # "Script"
        inst.setStringPref("params", "")      # "Script Arguments"
        inst.setStringPref("cwd", "")         # "Directory"
        inst.setBooleanPref("show-dialog", 1) 
        inst.setStringPref("language", self._language) 

        inst.parent = cat
        self._getSystemEnvironment(cat, inst)
        self._getDefaultCGIPrefs(cat, inst)
        self._fillExtraDefaultPreferences(cat, inst)

        return cat, inst

    def get_currentCategoryPreferences( self ):
        # Result: koIPreferenceSet
        return self.currentCategoryPrefs

    def set_currentCategoryPreferences( self, prefs):
        # Result: void - None
        # In: param0: koIPreferenceSet
        self.currentCategoryPrefs = prefs

    def get_currentInstancePreferences( self ):
        # Result: koIPreferenceSet
        return self.currentInstancePrefs

    def set_currentInstancePreferences( self, prefs):
        # Result: void - None
        # In: param0: koIPreferenceSet
        if prefs.hasStringPref("language") and prefs.getStringPref("language") != self._language:
            raise ServerException(nsError.NS_ERROR_UNEXPECTED, "Invocation Preference language missmatch")
        self.currentInstancePrefs = prefs

    def get_persistableInstancePreferences( self ):
        # Result: koIPreferenceSet
        ret = self.currentInstancePrefs.clone()
        for prefName in ["CGIEnvironment", "systemEnvironment"]:
            if ret.hasPref(prefName):
                ret.deletePref(prefName)
        return ret

    def _doLaunch(self, argv, cwd, env):
        showConsole = self.currentInstancePrefs.getBooleanPref('use-console')
        scriptFileName = None
        inputFileName = None
        try:
            if showConsole:
                scriptFileName, inputFileName = self._doLaunchInConsole(argv, cwd, env)
                #self.process = process.Process(argv, cwd, env,
                #                               process.Process.CREATE_NEW_CONSOLE)
            else:
                # if we're doing cgi emulation, we ignore the stdin
                # from the terminal view.  POST data goes out over
                # standard process handles, and NOT over our redirected
                # stdin/socket stuff in koRequesterDebugger.py
                if self._postData and self.stdin:
                    self.stdin = None
                
                self.process = process.ProcessProxy(argv, mode='b', cwd=cwd,
                                                    env=env,
                                                    stdin=self.stdin,
                                                    stdout=self.stdout,
                                                    stderr=self.stderr)
                
                if self._postData and self.process.stdin:
                    self.process.stdin.write(self._postData)
                    self.process.stdin.close();
        except process.ProcessError, ex:
            msg = "Executing '%s' failed: %s\n" % (argv, str(ex))
            log.warn(msg)
            raise ServerException(nsError.NS_ERROR_FILE_EXECUTION_FAILED, msg)

        # The "invocation_terminated" notification is sent when the
        # child terminates.  A separate thread is created to handle that
        # so this call can return immediately.
        t = threading.Thread(target=self._WaitCleanUpAndNotify,
                             kwargs={'scriptFileName': scriptFileName,
                                     'inputFileName': inputFileName})
        t.start()

    def _doLaunchInConsole(self, command, cwd, envDict):
        """Launch the given command string in a new console.
        (NOTE: this is mostly copied from koRunService.Run())

        Optionally an "input" string can be provided. If it is then the
        command is run via "command < input", where the input is placed
        in a temporary file.
        """
        #print "_doLaunchInConsole(command=%r, cwd=%r, env=%r)"\
        #      % (command, cwd, envDict)

        # Keep a pure 'command' for reporting status to the user.
        actualCommand = _joinArgv(command)

        if self._postData:
            # Put the input into a temp file.
            inputFileName = tempfile.mktemp()
            inputFile = open(inputFileName, "w")
            inputFile.write(self._postData)
            inputFile.close()
            if ' ' in inputFileName:
                actualCommand += ' < "%s"' % inputFileName
            else:
                actualCommand += ' < %s' % inputFileName
        else:
            inputFileName = None

        if sys.platform.startswith("win"):
            if os.environ.has_key("SHELL"):
                shell = os.environ["SHELL"]
            elif os.environ.has_key("ComSpec"):
                shell = os.environ["ComSpec"]
            elif isWin9x:
                #XXX Would be nice to get full path here so can use
                #    os.spawn* calls.
                shell = "command.com"
            else:
                shell = "cmd.exe"
            shellArgs = "/C"
        elif sys.platform == "darwin":
            shell = "/bin/sh"
            shellArgs = ""
        else:
            shell = "xterm"
            shellArgs = "-e /bin/sh"
    
        # A shell script that will be run is created.  The script
        # adds the "Press any key to continue..." feature and
        # ensures that the correct return value is returned.
        if sys.platform.startswith("win"):
            scriptFileName = tempfile.mktemp() + ".bat"
            bat = open(scriptFileName, "w")
            bat.write("@echo off\n")
            # Escape '%' once for being in a batch file plus once for
            # being on a "call ..." line.
            actualCommand = actualCommand.replace("%", "%%%%")
            bat.write("call %s\n" % actualCommand)
            bat.write("set KOMODO_COMMAND_ERRORLEVEL=%ERRORLEVEL%\n")
            bat.write("pause\n")
            # Must "exit" with the last return value to ensure that
            # cmd.exe will return that value
            #XXX does the correct retval actually get returned?
            bat.write("exit %KOMODO_COMMAND_ERRORLEVEL%\n")
            bat.close()
            #XXX does this work on Win9x???
        elif sys.platform == "darwin":
            #XXX This code is (mostly) a duplicate of that in
            #    koRunService.py. They should be shared.
            #
            # We can use applescript to run a shell command in a new
            # Terminal.app window. The problem is that the relevant
            # AppleScript command does not wait for the command to
            # terminate. We need to resort to locks and temp files to
            # (a) wait for termination and (b) get the retval.
            #
            # 1. lockfile foo.lock
            # 2. run foo.applescript to run the following in the terminal:
            #    - run the user's command
            #    - write the retval to a temp file
            #    - rm -f foo.lock
            # 3. lockfile -1 foo.lock  # i.e. wait for foo.lock to get rm'd
            # 4. exit with the retval
            #
            def escapecmd(cmd):
                return cmd.replace('"', '\\"').replace('$', '\\$')
            base = tempfile.mktemp()
            shEnvFile = base+".environ.sh"
            cshEnvFile = base+".environ.csh"
            data = {
                "lockfile": base+".lock",
                "retvalfile": base+".retval",
                "applescriptfile": base+".applescript",
                "cmd": escapecmd(actualCommand),
                "cwdcmd": "",
                "shenvfile": shEnvFile,
                "cshenvfile": cshEnvFile,
                "envcmd": (r"echo \$shell | grep csh >/dev/null "
                           r" && source %s "
                           r" || source %s" % (cshEnvFile, shEnvFile)),
            }
            if cwd:
                data["cwdcmd"] = r'cd \"%s\";' % cwd

            def quotevalue(value):
                if "'" in value:
                    for ch in "\\$!": # special shell chars to escape
                        value = value.replace(ch, '\\'+ch)
                    return '"%s"' % value
                else:
                    return "'%s'" % value
            fenv = open(shEnvFile, "w")
            for name, value in envDict.items():
                fenv.write('    %s=%s; export %s\n' % (name, quotevalue(value), name))
            fenv.close()
            fenv = open(cshEnvFile, "w")
            for name, value in envDict.items():
                fenv.write('    setenv %s %s\n' % (name, quotevalue(value)))
            fenv.close()

            scriptFileName = base+".sh"
            script = open(scriptFileName, "w")
            script.write(r"""
LOCKFILE='%(lockfile)s'
RETVALFILE='%(retvalfile)s'
APPLESCRIPTFILE='%(applescriptfile)s'
SHENVFILE='%(shenvfile)s'
CSHENVFILE='%(cshenvfile)s'

rm -f $LOCKFILE
lockfile $LOCKFILE
cat >$APPLESCRIPTFILE <<HERE
tell app "Terminal"
    do script "clear; %(cwdcmd)s (%(envcmd)s; %(cmd)s); echo \$?>$RETVALFILE; echo Press RETURN to continue ...; /bin/sh -c read RETURN; rm -f $LOCKFILE; exit"
    delay 0.2
    activate
end tell
HERE
#echo debug: running AppleScript in $APPLESCRIPTFILE:
#cat $APPLESCRIPTFILE
#echo debug: which will source one of $SHENVFILE:
#cat $SHENVFILE
#echo debug: or $CSHENVFILE:
#cat $CSHENVFILE
osascript $APPLESCRIPTFILE
#echo debug: waiting for Terminal to finish...
lockfile -1 $LOCKFILE
RETVAL=`cat $RETVALFILE`
#echo debug: retval was \"$RETVAL\"
rm -f $RETVALFILE $LOCKFILE $SHENVFILE $CSHENVFILE $APPLESCRIPTFILE
#XXX Need Komodo to register with OSA (bug 37377).
#osascript -e 'tell app "Komodo" to activate'
exit $RETVAL""" % data)
            script.close()
        else:
            # Create a bash script that will be run.
            scriptFileName = tempfile.mktemp() + ".sh"
            script = open(scriptFileName, "w")
            script.write("#!/bin/sh\n")
            script.write("%s\n" % actualCommand)
            script.write("export KOMODO_COMMAND_RETVAL=$?\n")
            script.write("echo -n Press ENTER to continue . . .\n")
            script.write("read KOMODO_DUMMY\n")
            script.write("unset KOMODO_DUMMY\n")
            script.write("exit $KOMODO_COMMAND_RETVAL\n")
            script.close()
        actualCommand = "%s %s %s"\
                        % (shell, shellArgs, scriptFileName)
        #print "RUN: actualCommand is '%s'" % actualCommand
        #print "RUN: -------- %s is as follows -------" % scriptFileName
        #fin = open(scriptFileName, 'r')
        #for line in fin.readlines():
        #    print line,
        #fin.close()
        #print "RUN: ---------------------------------"

        try:
            self.process = process.Process(actualCommand, cwd=cwd, env=envDict,
                                    flags=process.Process.CREATE_NEW_CONSOLE)
        except process.ProcessError, ex:
            log.exception(ex)
            raise ServerException(nsError.NS_ERROR_FAILURE, str(ex))
        
        return scriptFileName,inputFileName

    def _WaitCleanUpAndNotify(self, scriptFileName=None,
                                    inputFileName=None):
        """Thread used by _doLaunchInConsole."""
        retval = self.process.wait()

        if scriptFileName:
            try:
                os.unlink(scriptFileName)
            except OSError, ex:
                log.warn("Could not remove temporary script file '%s': %s",
                         scriptFileName, ex)
        if inputFileName:
            try:
                os.unlink(inputFileName)
            except OSError, ex:
                log.warn("Could not remove temporary input file '%s': %s",
                         inputFileName, ex)
        
        # Send appropriate notification.
        try:
            #XXX Should eventually pass exitCode out with notification
            self._observerSvc.notifyObservers(self, "invocation_terminated", None)
        except COMException, e:
            # do nothing if nobody listening
            pass

    def _setPrefEnvironment(self, inst, name, newname = None):
        if not newname: newname = name
        penv = self.prefService.prefs.getStringPref(name)
        # if we have a predefined user environment, sort it first
        if len(penv) > 0:
            penv = string.split(penv, os.linesep)
            penv.sort()
            penv = string.join(penv,os.linesep)
        inst.setStringPref(newname,penv)
        
    def _getSystemEnvironment(self, cat, inst):
        environment = components.classes["@activestate.com/koUserEnviron;1"].getService().GetEnvironmentStrings()
        environment.sort()
        env = string.join(environment,"\n")
        cat.setStringPref("systemEnvironment",env)
        self._setPrefEnvironment(inst, "userEnvironment")
        
    # CGI emulation support functions
    def _getDefaultCGIPrefs(self, cat, inst):
        inst.setBooleanPref("sim-cgi",0)

        # get the webserver/browser env vars
        self._setPrefEnvironment(cat, "debuggerEnvironment", "CGIEnvironment")
        self._setPrefEnvironment(inst, "userCGIEnvironment")
        
        inst.setStringPref("documentRoot","")
        
        inst.setStringPref("getparams","")
        inst.setStringPref("postparams","")
        inst.setStringPref("posttype","application/x-www-form-urlencoded")
        inst.setStringPref("mpostparams","")
        inst.setStringPref("cookieparams","")
        inst.setStringPref("request-method","GET")
        
    def _getenv(self):
        """Return the merged list of environment strings.
        The merge order is:
            1. the system environment
            2. the default CGI environment
            3. the user environment settings
            4. the user CGI environment settings
        This tries to ensure, as best as possible, that the user's
        settings are used as expected.
        """
        # 1. the system environment
        mergedEnvStrs = self.currentInstancePrefs.getStringPref('systemEnvironment')
        mergedEnvStrs = mergedEnvStrs.strip().split("\n")

        # 2. the default CGI environment
        if self._usecgi:
            cgiEnvStrs = self.currentInstancePrefs.getStringPref('CGIEnvironment')
            cgiEnvStrs = cgiEnvStrs.strip().split("\n")
            mergedEnvStrs = self._environUtils.MergeEnvironmentStrings(
                mergedEnvStrs, cgiEnvStrs)
            
        # 3. the user environment settings
        userEnvStrs = self.currentInstancePrefs.getStringPref('userEnvironment')
        if userEnvStrs:
            userEnvStrs = userEnvStrs.strip().split("\n")
            mergedEnvStrs = self._environUtils.MergeEnvironmentStrings(
                mergedEnvStrs, userEnvStrs)
        
        # 4. the user CGI environment settings
        if self._usecgi:
            userCgiEnvStrs = self.currentInstancePrefs.getStringPref('userCGIEnvironment')
            if userCgiEnvStrs:
                userCgiEnvStrs = userCgiEnvStrs.strip().split("\n")
                mergedEnvStrs = self._environUtils.MergeEnvironmentStrings(
                    mergedEnvStrs, userCgiEnvStrs)
            
            # add generated cgi env vars (from GET, POST, and Cookie params)
            httpEnvStrs = ["%s=%s" % item
                           for item in self._build_http_env().items()]
            mergedEnvStrs = self._environUtils.MergeEnvironmentStrings(
                mergedEnvStrs, httpEnvStrs)

        # Invariant:
        #   The MergeEnvironmentStrings() functionality ensures that
        #   there are no env. vars with an empty value.
        # Convert items to plain strings (no Unicode)
        mergedEnvStrs = [str(item) for item in mergedEnvStrs]
        # add komodo version
        mergedEnvStrs.append("KOMODO_VERSION=%s" % self.komodoVersion)
        mergedEnvStrs.append("DBGP_COOKIE=%s" % self.cookie)

        return environmentStringsToDict(mergedEnvStrs)
    
    def _filename_from_url(self, url):
        fn = re.match(r'.*[\/\\]([^\/\\]*)$',url);
        if fn:
            return fn.group(1)
        return ""
        
    def _build_http_env(self):
        self._postData, self._postSize, self._postType = self._build_post_data()
        requestMethod = self.currentInstancePrefs.getStringPref('request-method')
        
        # calculate the document root
        pathTranslated = self.currentInstancePrefs.getStringPref('filename')
        cookieData = self.currentInstancePrefs.getStringPref('cookieparams')
        cookies = self._build_args_list(cookieData, "; ")
        
        # XXX have to do something MUCH better for the doc root!
        scriptName = "/" + self._filename_from_url(pathTranslated)

        env = {}
        env["QUERY_STRING"]=self._build_get_line()
        env["CONTENT_LENGTH"]=self._postSize
        if self._postSize:
            #print "Content size is ",self._postSize
            env["CONTENT_TYPE"]=self._postType
        env["PATH_INFO"]=scriptName #XXX verify this!
        env["PATH_TRANSLATED"]=pathTranslated
        env["SCRIPT_NAME"]=scriptName
        env["REQUEST_METHOD"]=requestMethod
        env["HTTP_COOKIE"]=cookies
        # build any environment variables we need to emulate the web server environment
        return env
    
    def _build_args_list(self, args, sep="&"):
        argsList = string.split(string.strip(args), "\n")
        return string.join(argsList,sep)
    
    def _build_get_line(self):
        getArgs = self.currentInstancePrefs.getStringPref('getparams')
        return self._build_args_list(getArgs)

    def _build_post_data(self):
        # build the post data that must be sent down stdin to the cgi
        # XXX extremely simplistic mime building going on here

        postArgs = self.currentInstancePrefs.getStringPref('postparams')
        postFiles = self.currentInstancePrefs.getStringPref('mpostparams')
        postType = self.currentInstancePrefs.getStringPref('posttype')
        
        if not postType: postType = "application/x-unknown"
        
        if postType == "application/x-www-form-urlencoded":
            postData = self._build_args_list(postArgs)
        elif postType == "multipart/form-data":
            # handle RFC 1867 data
            out = ""
            # we have to do multpart mime encoding
            boundary = mimetools.choose_boundary()
            # generate the mime headers for regular post variables
            if postArgs:
                argsList = string.split(postArgs, "\n")
                for i in range(len(argsList)):
                    try:
                        key, value = argsList[i].split("=", 1)
                        out += "--"+boundary+"\r\n"
                        out += "Content-Disposition: form-data; name=\"%s\"\r\n\r\n" % (key)
                        out += value+"\r\n"
                    except ValueError:
                        pass
                
            if postFiles:
                fileList = string.split(postFiles, "\n")
                # read each file, encode it, then add it to the return data
                for i in range(len(fileList)):
                    # read each file, base64 encode it, and add to the request
                    if len(fileList[i]) < 3: continue
                    fvar = string.split(fileList[i], "=")
                    fn = self._filename_from_url(fvar[1])
                    out += "--"+boundary+"\r\n"
                    out += "Content-Disposition: form-data; name=\"%s\"; filename=\"%s\"\r\n" % (fvar[0], fn)
                    type, encoding = mimetypes.guess_type(fvar[1])
                    if not type: type = "text/plain"
                    out += "Content-Type: %s\r\n" % (type)
                    out += "Content-Transfer-Encoding: binary\r\n\r\n"

                    infile = open(fvar[1], "rb")
                    out += infile.read()
                    infile.close()
                    
            # XXX - not rfc compliant, but works right, should have additional -- at end
            out += "--"+boundary+"\r\n"
            postData = out
            postType = "multipart/form-data; boundary=\"%s\"" % (boundary)
        
        else:
            # support raw post data entered into the post args field
            postData = postArgs
        return postData,len(postData),postType

class koProjectInvocation(koInterpreterInvocation):
       # If this object needs to be registered, the following 2 are also needed.
    _language = ""
    _reg_clsid_ = "{25ed901e-4ffe-11db-82f2-000d935d3368}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=Project"
    #_reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "Project"
    _invocation_description_ = "Project"

    def validInterpreter(self, invokeType):
        # we dont validate at the project pref level
        return 1, None, None

    def set_currentInstancePreferences( self, prefs):
        # Result: void - None
        # In: param0: koIPreferenceSet
        prefs.setStringPref("langauge", self._language) 
        self.currentInstancePrefs = prefs

class koPythonInvocation(koInterpreterInvocation):
       # If this object needs to be registered, the following 2 are also needed.
    _language = "Python"
    _reg_clsid_ = "{eef242ad-c57e-4507-80d5-d2fa7885dea5}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=Python"
    #_reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "Python"
    _invocation_description_ = "Python"

    def _updatePythonPath(self, prefset, env):
        if prefset.hasPref("pythonExtraPaths"):
            pythonPath = prefset.getStringPref("pythonExtraPaths")
        if pythonPath:
            pythonPathEnv = env.get("PYTHONPATH", "")
            if pythonPathEnv:
                pythonPath += os.pathsep + pythonPathEnv

        if pythonPath:
            if sys.platform.startswith("win"):
                pythonPath = string.replace(pythonPath, '\\', '/')
            env['PYTHONPATH'] = pythonPath
    
    def _getSystemEnvironment(self, cat, inst):
        environment = components.classes["@activestate.com/koUserEnviron;1"].getService().GetEnvironmentStrings()
        environment = environmentStringsToDict(environment)
        self._updatePythonPath(self.prefService.prefs, environment)
        environment = environmentDictToStrings(environment)
        environment.sort()
        env = string.join(environment,"\n")
        cat.setStringPref("systemEnvironment",env)
        self._setPrefEnvironment(inst, "userEnvironment")

    def _fillExtraDefaultPreferences(self, cat, inst):
        cat.setBooleanPref("unbuffered-handles", 1)

    def _determineExecutable(self):
        registryService = components.classes['@activestate.com/koLanguageRegistryService;1'].\
               getService(components.interfaces.koILanguageRegistryService)
        languageService = registryService.getLanguage('Python');

        pythonExe = self.prefService.prefs.getStringPref("pythonDefaultInterpreter")
        if not pythonExe:
            pythonExe = languageService.getLanguageService(components.interfaces.koIAppInfoEx).executablePath
        if not pythonExe:
            log.error("No Python interpreter could be found.")
            return ""
        else:
            return pythonExe

    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the executable used for the debug session")
            if invokeType != INVOKE_INTERACTIVE:
                checkFileExists(prefset, 'filename', "You must specify the main Python script to be used")

            cwd = prefset.getStringPref("cwd")
            if cwd and not os.path.isdir(cwd):
                raise PrefException("cwd", "Directory '%s' does not exist" % cwd)
        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None

    def _getExecutableArgs(self, executable):
        cmddata = self.currentInstancePrefs.getStringPref('executable-params')
        executableArgs = listcmd.line2argv(cmddata)
        if '-u' not in executableArgs and \
           os.path.basename(executable).find('jython') == -1 and \
           self.currentInstancePrefs.getBooleanPref('unbuffered-handles'):
            executableArgs.append("-u")
        return executableArgs
    
    def _getDBGPClientDriverArgs(self):
        driverArgs = []
        port = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        driverArgs.append("-d")
        driverArgs.append("127.0.0.1:%s" % port)
        return driverArgs
    
    def _getDBGPClientDriver(self, executable):
        if sys.platform == "win32":
            driverBase = "pydbgp.py"
        else:
            driverBase = "pydbgp"
        driver = os.path.join(os.path.dirname(executable), driverBase)
        if os.path.exists(driver):
            return driver, 1

        # The Python interpreter that is used to launch "dbgp" may be a
        # a different version than is currently run, therefore we must be
        # sure to launch pydbpg.py rather pydbgp.py[oc].
        koDirs = components.classes["@activestate.com/koDirs;1"].\
                 getService(components.interfaces.koIDirs)
        return os.path.join(koDirs.binDBGPDir, driverBase), 0
    
    def _getScriptArgs(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')
        # use the get variables as the user_args if emulating CGI        
        scriptArgs = []
        if self._usecgi:
            get_args = self._build_get_line()
            if get_args: scriptArgs.append(get_args)
        if len(scriptArgs) < 1:
            cmddata = self.currentInstancePrefs.getStringPref('params')
            scriptArgs = listcmd.line2argv(cmddata)
        return scriptArgs
        
    def _getDebuggerEnvironment(self, installed=0):
        # Extend the environment so our dbgp Python package is used.
        env = self._getenv()

        koDirSvc = components.classes["@activestate.com/koDirs;1"].\
                   getService(components.interfaces.koIDirs)

        # In a dev build environment there are some dirs on the PYTHONPATH
        # that will get in the way. In particular the siloed Python
        # "site-packages" has a 'dbgp' package withOUT the DBGP client bits
        # (wich would cause 'pydbgp' to fail when importing dbgp.client) and
        # the PyXPCOM import directory has stuff that isn't needed.
        currPyPath = env.get("PYTHONPATH", "").split(os.pathsep)
        cleanedPyPath = []
        baddies = [
            os.path.join(koDirSvc.mozBinDir, "python"),
            # Don't want siloed Python "site-packages" dir. Cheat: assume any
            # path starting with Komodo's install dir.
            koDirSvc.installDir,
        ]
        for path in currPyPath:
            for baddy in baddies:
                if path.startswith(baddy):
                    break
            else:
                # Okay to keep this PYTHONPATH entry.
                cleanedPyPath.append(path)
        cleanedPyPath = os.pathsep.join(cleanedPyPath)
        env["PYTHONPATH"] = cleanedPyPath

        # set the dbgp site-packages into PYDBGP_PATH
        # using this allows python -E to work
        if not installed:
            dbgpLibDir = koDirSvc.pythonDBGPDir
            if not os.path.exists(dbgpLibDir):
                errmsg = "cannot debug: cannot find Python DBGP support "\
                         "libs: '%s'" % dbgpLibDir
                log.error(errmsg)
                #XXX *Could* set last error on koILastErrorSvc but the invocation
                #    system doesn't use that yet.
                raise ServerException(nsError.NS_ERROR_UNEXPECTED, errmsg)
            env["PYDBGP_PATH"] = dbgpLibDir

        if log.isEnabledFor(logging.DEBUG):
            log.debug("%s Python Debugging Environment", "="*40)
            log.debug("PYTHONPATH=%s", env.get("PYTHONPATH", ""))
            log.debug("PYDBGP_PATH=%s", env.get("PYDBGP_PATH", ""))
            log.debug("="*70)
        return env

    def _launchForInteractive(self):
        executable = self.currentInstancePrefs.getStringPref('executable')
        executableArgs = self._getExecutableArgs(executable)
        driver, installed = self._getDBGPClientDriver(executable)
        driverArgs = self._getDBGPClientDriverArgs()
        if '-i' in executableArgs:
            executableArgs.remove('-i')
        argv = [executable] + executableArgs + [driver] + driverArgs + ['-i']

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.getcwd()
        env = self._getDebuggerEnvironment(installed)
        return self._doLaunch(argv, cwd, env)

    def _launchForDebug(self):
        executable = self.currentInstancePrefs.getStringPref('executable')
        executableArgs = self._getExecutableArgs(executable)
        driver, installed = self._getDBGPClientDriver(executable)
        driverArgs = self._getDBGPClientDriverArgs()
        if '-i' in executableArgs:
            executableArgs.remove('-i')
            driverArgs.append('-i')
        script = self.currentInstancePrefs.getStringPref('filename')
        scriptArgs = self._getScriptArgs()
        argv = [executable] + executableArgs + [driver] + driverArgs \
               + [script] + scriptArgs

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        env = self._getDebuggerEnvironment(installed)
        return self._doLaunch(argv, cwd, env)

    def _launchForRun(self):
        executable = self.currentInstancePrefs.getStringPref('executable')
        executableArgs = self._getExecutableArgs(executable)
        driver, installed = self._getDBGPClientDriver(executable)
        driverArgs = self._getDBGPClientDriverArgs()
        if '-i' in executableArgs:
            executableArgs.remove('-i')
            driverArgs.append('-i')
        driverArgs.append('-n') # run without debug
        script = self.currentInstancePrefs.getStringPref('filename')
        scriptArgs = self._getScriptArgs()
        argv = [executable] + executableArgs + [driver] + driverArgs \
               + [script] + scriptArgs

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        env = self._getDebuggerEnvironment(installed)
        return self._doLaunch(argv, cwd, env)

class koPerlInvocation(koInterpreterInvocation):
    _language = "Perl"
    _reg_clsid_ = "{3becc3bd-1e42-47b4-aafd-9974d10823dd}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=Perl"
    #_reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "Perl"
    _invocation_description_ = "Perl"

    def __init__(self):
        koInterpreterInvocation.__init__(self)
        proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerSvc = proxyMgr.getProxyForObject(None,
            components.interfaces.nsIObserverService,
            observerSvc, PROXY_ALWAYS | PROXY_SYNC)
        registryService = components.classes['@activestate.com/koLanguageRegistryService;1'].\
               getService(components.interfaces.koILanguageRegistryService)
        self.languageService = registryService.getLanguage('Perl');
    
    def _fillExtraDefaultPreferences(self, cat, inst):
        inst.setStringPref("warnings", "enabled")
        inst.setLongPref("debugger.io-port", 9011)

    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the Perl executable used for the debug session")
            if invokeType != INVOKE_INTERACTIVE:
                checkFileExists(prefset, 'filename', "You must specify the main Perl script to be used")

            cwd = prefset.getStringPref("cwd")
            if cwd and not os.path.isdir(cwd):
                raise PrefException("cwd", "Directory '%s' does not exist" % cwd)
            warnings = prefset.getStringPref("warnings")
            valid_warning_values = ["enabled", "disabled", "all"]
            if warnings not in valid_warning_values:
                raise PrefException("warnings", "Must be one of " + ",".join(valid_warning_values))
        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None
        
    def _determineExecutable(self):
        # use pref'd default if one has been chosen
        perlInfo  = self.languageService.\
                   getLanguageService(components.interfaces.koIAppInfoEx)
        perlExe = perlInfo.executablePath
        if not perlExe:
            installations = perlInfo.FindInstallationPaths()
            if installations:
                perlExe = installations[0]
        if not perlExe or not os.path.isfile(perlExe):
            log.error("No Perl interpreter could be found.")
            return ""
        return perlExe


    def _getExtraPerlOptions(self):
        cmddata = self.currentInstancePrefs.getStringPref('executable-params')
        ret = listcmd.line2argv(cmddata)

        if '-w' not in ret and '-W' not in ret and '-X' not in ret:
            warnings = self.currentInstancePrefs.getStringPref('warnings')
            if warnings == "enabled":
                ret.append("-w")
            elif warnings == "all":
                ret.append("-W")
            elif warnings == "disabled":
                ret.append("-X")
        return ret
    
    def _getPerlIncArgv(self):
        argv = []
        perlExtraPaths = []
        if self.prefService.prefs.hasPref("perlExtraPaths"):
            perlExtraPaths = self.prefService.prefs.getStringPref("perlExtraPaths")
        if perlExtraPaths:
            if sys.platform.startswith("win"):
                perlExtraPaths = string.replace(perlExtraPaths, '\\', '/')
            perlExtraPaths = [x for x in perlExtraPaths.split(os.pathsep) if x.strip()]
            for incpath in perlExtraPaths:
                argv += ['-I', incpath]
        return argv

    def _getPerlDbgLibForCmdLine(self):
        # Bug OTS#: 10101113-FW -- In taint mode Perl doesn't
        # consult the PERL5LIB or PERLLIB env vars, but it
        # will honor -I dirs.
        
        koDirSvc = components.classes["@activestate.com/koDirs;1"].\
                        getService(components.interfaces.koIDirs)
        dbgpLibDir = koDirSvc.perlDBGPDir
        if not os.path.exists(dbgpLibDir):
            return []
        return ['-I', dbgpLibDir]

    def _getLogFile(self):
        # The PERLDB_OPTS LogFile setting is given only if
        # logEnabled and len(logPath/logDir) > 0 (duh)
        # add logging if configured
        logEnabled = 0
        logDir = ''
        if self.prefService.prefs.hasPref("perl_debuggerlogenabled"):
            logEnabled = self.prefService.prefs.getBooleanPref("perl_debuggerlogenabled")
        if logEnabled:
            if self.prefService.prefs.hasPref("perl_debuggerlogpath"):
                logDir = self.prefService.prefs.getStringPref("perl_debuggerlogpath")
                logDir = os.path.expanduser(logDir)
            if not logDir:
                logDir = os.getenv("TEMP", os.getenv("TMP", ''))
                if logDir:
                    logDir = os.path.join(logDir, 'perl-dbgp.log')
            if logDir:
                logDir = logDir.replace("\\", "/")
                # encode the path
                logDir = urllib.quote(logDir)
        return logDir
    
    def _getScriptArgs(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')

        # use the get variables as the user_args if emulating CGI        
        scriptArgs = []
        if self._usecgi:
            get_args = self._build_get_line()
            if get_args: scriptArgs.append(get_args)
        if len(scriptArgs) < 1 and \
           self.currentInstancePrefs.hasPref('params'):
            params = self.currentInstancePrefs.getStringPref('params')
            if params:
                scriptArgs = listcmd.line2argv(params)
        return scriptArgs
    
    def _getDebuggerEnvironment(self):
        # Setup the environment so that our perl5db.pl (and the extra
        # Perl module it requires) get used.
        env = self._getenv()

        cmdPort = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        #listenerPort = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        envOpts = env.get("PERLDB_OPTS", "").split(os.pathsep)
        # Keep the old env settings

        perldbOpts = {}
        for pair in [opt.split("=", 1) for opt in envOpts if opt]:
            if len(pair) == 1:
                perldbOpts[pair[0]] = None
            else:
                perldbOpts[pair[0]] = pair[1]
        
        # Override new ones
        perldbOpts['RemotePort'] = '127.0.0.1:%d' % (cmdPort) # Required
        logFile = self._getLogFile()
        if logFile:
            perldbOpts['LogFile'] = logFile

        # Recreate the env var
        env["PERLDB_OPTS"] = ' '.join([(val is not None and (key + '=' + val) or key)
                                       for key,val in perldbOpts.items()])

        #XXX Is this block correct for Linux/Solaris?
        koDirSvc = components.classes["@activestate.com/koDirs;1"].\
                   getService(components.interfaces.koIDirs)
        perl5lib = env.get("PERL5LIB", "").split(os.pathsep)
        dbgpLibDir = koDirSvc.perlDBGPDir
        try:
            log.debug("dbgpLibDir = %s", dbgpLibDir);
        except:
            log.debug("no val for dbgpLibDir");
        if not os.path.exists(dbgpLibDir):
            errmsg = "cannot debug: cannot find Perl DBGP support libs: '%s'" % dbgpLibDir
            log.error(errmsg)
            #XXX *Could* set last error on koILastErrorSvc but the invocation
            #    system doesn't use that yet.
            raise ServerException(nsError.NS_ERROR_UNEXPECTED, errmsg)
        perl5lib.insert(0, dbgpLibDir)
        perl5lib = os.pathsep.join(perl5lib)
        try:
            log.debug("perl5lib = %s", perl5lib);
        except:
            log.debug("no env val for perl5lib");
        env["PERL5LIB"] = perl5lib
        perl5db = "BEGIN { require '%s' }"\
                         % os.path.join(dbgpLibDir, 'perl5db.pl')
        if sys.platform.startswith("win"):
            env["PERL5DB"] = string.replace(perl5db, '\\', '/')
        else:
            env["PERL5DB"] = perl5db
        if log.isEnabledFor(logging.DEBUG):
            log.debug("%s Perl Debugging Environment", "="*40)
            log.debug("PERL5LIB=%s", env['PERL5LIB'])
            log.debug("PERL5DB=%s", env['PERL5DB'])
            log.debug("PERLDB_OPTS=%s", env['PERLDB_OPTS'])
            log.debug("="*70)
        return env

    def _launchForInteractive(self):
        executable = self.currentInstancePrefs.getStringPref('executable')
        argv = [executable, "-d", "-e", "1"] \
               + self._getExtraPerlOptions() \
               + self._getPerlDbgLibForCmdLine() \
               + self._getPerlIncArgv()
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.getcwd()
        env = self._getDebuggerEnvironment()
        self._doLaunch(argv, cwd, env)
    
    def _launchForDebug(self):
        script = self.currentInstancePrefs.getStringPref('filename')
        executable = self.currentInstancePrefs.getStringPref('executable')
        argv = [executable, "-d"] \
               + self._getExtraPerlOptions() \
               + self._getPerlDbgLibForCmdLine() \
               + self._getPerlIncArgv() + [script] \
               + self._getScriptArgs()
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        env = self._getDebuggerEnvironment()
        self._doLaunch(argv, cwd, env)

    def _launchForRun(self):
        script = self.currentInstancePrefs.getStringPref('filename')
        executable = self.currentInstancePrefs.getStringPref('executable')
        argv = [executable] \
               + self._getExtraPerlOptions() \
               + self._getPerlIncArgv() + [script] \
               + self._getScriptArgs()
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        env = self._getenv()
        self._doLaunch(argv, cwd, env)

class koPHPInvocation(koInterpreterInvocation):
    _language = "PHP"
    _reg_clsid_ = "{aa80dab8-b5fb-45c4-b1c5-0eec9494955b}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=PHP"
    #_reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "PHP"
    _invocation_description_ = "PHP"

    def __init__(self):
        koInterpreterInvocation.__init__(self)
        self.checked_can_debug = 0
        self.can_debug = 0
        registryService = components.classes['@activestate.com/koLanguageRegistryService;1'].\
               getService(components.interfaces.koILanguageRegistryService)
        self.languageService = registryService.getLanguage('PHP');
        self.error_invocation_not_supported = self._invocation_name_ + " is not configured for '%s'."

    def get_supportedInvokeTypes(self):
        rc = INVOKE_RUN
        self.can_debug = self.languageService.\
               getLanguageService(components.interfaces.koIAppInfoEx).\
               autoConfigureDebugger() == ""

        if self.can_debug:
            rc |= INVOKE_DEBUG
        return rc

    def _fillExtraDefaultPreferences(self, cat, inst):
        inst.setBooleanPref("disableOutputBuffering", 1)
        inst.setBooleanPref("enableImplicitFlush", 1)
        phpInfo = self.languageService.\
           getLanguageService(components.interfaces.koIAppInfoEx)
        if phpInfo.sapi == "cli":
            inst.setStringPref("phpInterpreterType", "php-cli")
        else:
            inst.setStringPref("phpInterpreterType", "php-cgi")
    
    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the PHP executable used for the debug session")
            checkFileExists(prefset, 'filename', "You must specify the main PHP script to be used")
            phpInfo = self.languageService.\
               getLanguageService(components.interfaces.koIAppInfoEx)
            if not phpInfo.valid_version:
                return 0, "", "Komodo does not support debugging with PHP Version %s, please upgrade to a later version." % phpInfo.version
            cwd = prefset.getStringPref("cwd")
            if cwd and not os.path.isdir(cwd):
                raise PrefException("cwd", "Directory '%s' does not exist" % cwd)
            
            #verify we can run
            if invokeType & INVOKE_DEBUG:
                rc = self.get_supportedInvokeTypes()
                if not rc & invokeType:
                    return 0, "", "PHP is not configured properly for debugging."
        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None
        
    def _determineExecutable(self):
        # use pref'd default if one has been chosen
        phpInterpreterType = self.currentInstancePrefs.getStringPref('phpInterpreterType')
        phpInfo = self.languageService.\
           getLanguageService(components.interfaces.koIAppInfoEx)
        if phpInterpreterType == 'php-cgi':
            phpExe = phpInfo.cgiExecutable
        else:
            phpExe = phpInfo.cliExecutable
        return phpExe

    def _launchForDebug(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')
        script = self.currentInstancePrefs.getStringPref('filename')
        ini = self.prefService.prefs.getStringPref("phpConfigFile")
        if not ini:
            phpInfo = self.languageService.\
               getLanguageService(components.interfaces.koIAppInfoEx)
            ini = phpInfo.cfg_file_path
        buffer = self.currentInstancePrefs.getBooleanPref("disableOutputBuffering")
        flush = self.currentInstancePrefs.getBooleanPref("enableImplicitFlush")
        if ini == "": ini = None
        executable = self.currentInstancePrefs.getStringPref('executable')

        # use the get variables as the user_args if emulating CGI        
        user_args = []
        if self._usecgi:
            get_args = self._build_get_line()
            if get_args: user_args.append(get_args)
        if len(user_args) < 1:
            params = self.currentInstancePrefs.getStringPref('params')
            user_args = listcmd.line2argv(params)

        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)
        
        exec_arg_string = ' '.join(execArgs)
        options = ['-d', 'cgi.force_redirect=off']
        if flush and exec_arg_string.find('implicit_flush') == -1:
            options += ['-d','implicit_flush=on']
        if buffer and exec_arg_string.find('output_buffering') == -1:
            options += ['-d','output_buffering=off']
            
        port = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        argv = [executable] + options + execArgs

        if self._usecgi:
            argv += [script]
        else:
            argv += ["-f", script] + user_args

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)

        environmentDict = self._getenv()
        environmentDict["XDEBUG_CONFIG"] = "remote_enable=1 remote_port=%d remote_handler=dbgp remote_mode=req" % port
        if ini is not None and '-c' not in argv:
            environmentDict["PHPRC"] = ini
        return self._doLaunch(argv, cwd, environmentDict)

    def _launchForRun(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')
        script = self.currentInstancePrefs.getStringPref('filename')
        buffer = self.currentInstancePrefs.getBooleanPref("disableOutputBuffering")
        flush = self.currentInstancePrefs.getBooleanPref("enableImplicitFlush")
        ini = self.prefService.prefs.getStringPref("phpConfigFile")
        if ini == "": ini = None
        executable = self.currentInstancePrefs.getStringPref('executable')
        params = self.currentInstancePrefs.getStringPref('params')
        user_args = listcmd.line2argv(params)

        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)

        exec_arg_string = ' '.join(execArgs)
        options = ['-d', 'cgi.force_redirect=off']
        if flush and exec_arg_string.find('implicit_flush') == -1:
            options += ['-d','implicit_flush=on']
        if buffer and exec_arg_string.find('output_buffering') == -1:
            options += ['-d','output_buffering=off']
            
        argv = [executable] + options + execArgs
        if self._usecgi:
            argv += [script]
        else:
            argv += ["-f", script] + user_args

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        environmentDict = self._getenv()
        environmentDict["XDEBUG_CONFIG"] = "remote_enable=off"
        if ini is not None and '-c' not in argv:
            environmentDict["PHPRC"] = ini
        
        return self._doLaunch(argv, cwd, environmentDict)




class koXSLTInvocation(koInterpreterInvocation):
    _language = "XSLT"
    _reg_clsid_ = "{20B8B708-D3A6-423b-BB41-9EDEE5604CD5}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=XSLT"
    #_reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "XSLT"
    _invocation_description_ = "XSLT"

    def __init__(self):
        koInterpreterInvocation.__init__(self)
        self.checked_can_debug = 0
        self.can_debug = 0

    def get_supportedInvokeTypes(self):
        return INVOKE_DEBUG | INVOKE_RUN

    def _checkValid(self, invokeType):
        return invokeType == INVOKE_DEBUG or invokeType == INVOKE_RUN
        
    def _fillExtraDefaultPreferences(self, cat, inst):
        inst.setStringPref("inputfile", "")

    def _getenv(self):     
        env = koInterpreterInvocation._getenv(self)
 
        if sys.platform.startswith("win"):
            pass
        elif sys.platform == "darwin":
            log.warn("XXX determine appropriate environment for XSLT invocation on Mac OS X")
            #XXX I suspect the right thing for Mac OS X is the
            #    following. Need to check, though.
            #koDirs = components.classes["@activestate.com/koDirs;1"]\
            #         .getService(components.interfaces.koIDirs)
            #env["DYLD_LIBRARY_PATH"] = os.path.join(koDirs.supportDir, "xslt")
        elif sys.platform.startswith("linux") \
             or sys.platform.startswith("solaris"):
            koDirs = components.classes["@activestate.com/koDirs;1"]\
                     .getService(components.interfaces.koIDirs)
            env["LD_LIBRARY_PATH"] = os.path.join(koDirs.supportDir, "xslt")
        return env
        
    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the XSLT executable used for the debug session")

            checkFileExists(prefset, 'filename', "You must specify the main XSLT script to be used")

            inputFile = prefset.getStringPref('inputfile')
            if not inputFile:
                raise PrefException('inputfile', "You must specify the input XML file to be used")
            else:
                addressingScheme, networkLocation, path, parameters, query, fragmentIdentifier = urlparse.urlparse(inputFile)
                if not addressingScheme:
                   if not os.path.isfile(inputFile):
                        raise PrefException('inputfile', "File '%s' does not exist" % inputFile)
                elif addressingScheme == 'file':
                    addressLocation = inputFile.lower().find('file:///')
                    if addressLocation == -1:
                        raise PrefException('inputfile', 'URIs that use the "file" addressing scheme must begin with "file:///"')
                    else:
                        if not os.path.isfile(inputFile[addressLocation + len('file:///'):]):
                            raise PrefException('inputfile', "File '%s' does not exist" % inputFile)

        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None
    
    def _determineExecutable(self):
        # use pref'd default if one has been chosen
        if self.prefService.prefs.hasStringPref("xsltDefaultInterpreter") and\
           self.prefService.prefs.getStringPref("xsltDefaultInterpreter"):
            runLocation = self.prefService.prefs.getStringPref("xsltDefaultInterpreter")
        else:        
            koDirs = components.classes["@activestate.com/koDirs;1"].\
                getService(components.interfaces.koIDirs)

            if sys.platform.startswith('win'): 
                runLocation = os.path.join(koDirs.supportDir, 'xslt', 'xsltdbgp.exe')
            else:
                runLocation = os.path.join(koDirs.supportDir, 'xslt', 'xsltdbgp')
            
        return runLocation

    def _launchForDebug(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')
        showConsole = 0
        script = self.currentInstancePrefs.getStringPref('filename')
        inputfile = self.currentInstancePrefs.getStringPref('inputfile')
        executable = self.currentInstancePrefs.getStringPref('executable')
        port = self.currentInstancePrefs.getLongPref("debugger.listener-port")

        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)

        argv = [executable] + execArgs + ["-dbgp", '127.0.0.1', str( port ), script, inputfile]
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        environment = self._getenv()
        return self._doLaunch(argv, cwd, environment)


    def _launchForRun(self):
        showConsole = 0
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')
        script = self.currentInstancePrefs.getStringPref('filename')
        inputfile = self.currentInstancePrefs.getStringPref('inputfile')
        executable = self.currentInstancePrefs.getStringPref('executable')

        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)

        argv = [executable] + execArgs + [script, inputfile]
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        return self._doLaunch(argv, cwd, self._getenv())



class koTclInvocation(koInterpreterInvocation):
       # If this object needs to be registered, the following 2 are also needed.
    _language = "Tcl"
    _reg_clsid_ = "{06860119-32f6-432d-a0c0-2398fcd56653}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=Tcl"
    #_reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "Tcl"
    _invocation_description_ = "Tcl"

    def _fillExtraDefaultPreferences(self, cat, inst):
        # XXX - It doesn't look like TCL_LIBRARYPATH is being used anywhere!
        cat.setStringPref("TCL_LIBRARYPATH", "")
        inst.setStringPref("tclInterpreterType", "wish")

    def _determineExecutable(self, type=None):
        """determine the Tcl interpreter to use"""
        # use pref'd default if one has been chosen

        # Find a reasonable Tcl installation (possible perf issue computing this
        # even when the user has picked an explicit preference)
        tclInstallation = components.classes["@activestate.com/koTclInfoEx;1"].createInstance();
        if not tclInstallation.selectDefault():
            tclInstallation = None
        
        if not type:
            type = self.currentInstancePrefs.getStringPref("tclInterpreterType")
        if type == "wish":
            tclExe = self.prefService.prefs.getStringPref("wishDefaultInterpreter")
            if not tclExe and tclInstallation:
                tclExe = tclInstallation.wish_path
        else:        
            tclExe = self.prefService.prefs.getStringPref("tclshDefaultInterpreter")
            if not tclExe and tclInstallation:
                tclExe = tclInstallation.tclsh_path
        if not tclExe and tclInstallation:
            self.currentInstancePrefs.setStringPref("tclInterpreterType", "tclsh")
            tclExe = tclInstallation.executablePath
        if not tclExe or not os.path.isfile(tclExe):
            log.error("No Tcl interpreter could be found.")
            return ""
        else:
            return tclExe
        
    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the executable used for the debug session")
            if invokeType != INVOKE_INTERACTIVE:
                checkFileExists(prefset, 'filename',   "You must specify the Tcl script to be debugged")

            cwd = prefset.getStringPref("cwd")
            if cwd and not os.path.isdir(cwd):
                raise PrefException("cwd", "Directory '%s' does not exist" % cwd)

            tclInfo = components.classes["@activestate.com/koTclInfoEx;1"] \
                     .getService(components.interfaces.koIAppInfoEx)
            tclInfo.installationPath = tclInfo.getInstallationPathFromBinary(prefset.getStringPref("executable"))

        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None

    def _getScriptArgs(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')

        # use the get variables as the user_args if emulating CGI        
        scriptArgs = []
        if self._usecgi:
            get_args = self._build_get_line()
            if get_args: scriptArgs.append(get_args)
        if len(scriptArgs) < 1 and \
           self.currentInstancePrefs.hasPref('params'):
            params = self.currentInstancePrefs.getStringPref('params')
            if params:
                scriptArgs = listcmd.line2argv(params)
        return scriptArgs
    
    def _getenv(self):
        environment = koInterpreterInvocation._getenv(self)
        
        # if there is no setting for the dev kit environment, use
        # the shared directory for it.  This enables site wide
        # use of *.pdx and *.pcx files for debugging
        if "TCLDEVKIT_LOCAL" not in environment:
            koDirs = components.classes["@activestate.com/koDirs;1"].\
                getService(components.interfaces.koIDirs)
            sharedDir = os.path.join(koDirs.commonDataDir, "tcl")
            environment["TCLDEVKIT_LOCAL"] = sharedDir

        if self.prefService.prefs.hasPref("tclExtraPaths"):
            tclExtraPaths = self.prefService.prefs.getStringPref("tclExtraPaths")
            # If TCLLIBPATH is set, then it must contain a valid Tcl
            # list giving directories to search during auto-load
            # operations. Directories must be specified in Tcl format,
            # using "/" as the path separator, regardless of platform.
            # This variable is only used when initializing the
            # auto_path variable.  Also escape spaces in paths.
            tclExtraPaths = tclExtraPaths.replace('\\', '/')
            tclExtraPaths = tclExtraPaths.replace(' ', '\ ')
            TCLLIBPATH = ' '.join(tclExtraPaths.split(os.pathsep))
            environment["TCLLIBPATH"] = TCLLIBPATH

        return environment
    
    def _getDBGPClientDriver(self):
        koDirs = components.classes["@activestate.com/koDirs;1"].\
            getService(components.interfaces.koIDirs)

        if sys.platform.startswith('win'):
            executable = 'dbgp_tcldebug.exe'
        else:
            executable = 'dbgp_tcldebug'

        return os.path.join(koDirs.supportDir, 'tcl', executable)

    def _getDBGPClientDriverArgs(self, type=None):
        interpreter = self._determineExecutable(type)
        port = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        driverArgs = ["-app-shell", interpreter,
                      "-host-ide", "127.0.0.1",
                      "-port-ide", str(port)]

        # add logging if configured
        logEnabled = 0
        logDir = ''
        if self.prefService.prefs.hasPref("tcl_debuggerlogenabled"):
            logEnabled = self.prefService.prefs.getBooleanPref("tcl_debuggerlogenabled")
        if logEnabled:
            if self.prefService.prefs.hasPref("tcl_debuggerlogpath"):
                logDir = self.prefService.prefs.getStringPref("tcl_debuggerlogpath")
                logDir = os.path.expanduser(logDir)
            if not logDir:
                logDir = os.getenv("TEMP", os.getenv("TMP", ''))
            if logDir:
                driverArgs += ["-log", "-logfile",
                               os.path.join(logDir, 'tcl.log')]
        return driverArgs

    def _launchForInteractive(self):
        # force tclsh for interactive shell
        driver = self._getDBGPClientDriver()
        driverArgs = self._getDBGPClientDriverArgs('tclsh') + ['-interactive']
        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)
        scriptArgs = self._getScriptArgs()
        argv = [driver] + execArgs + driverArgs + ['--'] + scriptArgs

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.getcwd()
        return self._doLaunch(argv, cwd, self._getenv())
    
    def _launchForDebug(self):
        script = self.currentInstancePrefs.getStringPref('filename')
        scriptArgs = self._getScriptArgs()
        driver = self._getDBGPClientDriver()
        driverArgs = self._getDBGPClientDriverArgs() + ['-app-file', script]
        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)
        argv = [driver] + execArgs + driverArgs + ['--'] + scriptArgs

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        return self._doLaunch(argv, cwd, self._getenv())

    def _launchForRun(self):
        executable = self._determineExecutable()
        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)
        script = self.currentInstancePrefs.getStringPref('filename')
        scriptArgs = self._getScriptArgs()
        argv = [executable] + execArgs + [script] + scriptArgs

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        return self._doLaunch(argv, cwd, self._getenv())

class koRubyInvocation(koInterpreterInvocation):
    _language = "Ruby"
    _reg_clsid_ = "{a580e9d6-8d99-4f65-b032-9062b5b9a3b0}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=Ruby"
    #_reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "Ruby"
    _invocation_description_ = "Ruby"
        
    def __init__(self):
        koInterpreterInvocation.__init__(self)
        self.checked_can_debug = 0
        self.can_debug = 0
        registryService = components.classes['@activestate.com/koLanguageRegistryService;1'].\
               getService(components.interfaces.koILanguageRegistryService)
        self.languageService = registryService.getLanguage('Ruby');
        koDirSvc = components.classes["@activestate.com/koDirs;1"].\
                   getService(components.interfaces.koIDirs)
        dbgpLibDir = os.path.join(koDirSvc.supportDir, 'dbgp', 'rubylib')
        if sys.platform.startswith('win'):
            self.dbgpLibDir = dbgpLibDir.replace('\\', '/')
        else:
            self.dbgpLibDir = dbgpLibDir

    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the Ruby executable used for the debug session")
            if invokeType != INVOKE_INTERACTIVE:
                checkFileExists(prefset, 'filename', "You must specify the main Ruby script to be used")

            cwd = prefset.getStringPref("cwd")
            if cwd and not os.path.isdir(cwd):
                raise PrefException("cwd", "Directory '%s' does not exist" % cwd)
        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None

    def _determineExecutable(self):
        # use pref'd default if one has been chosen
        rubyInfo  = self.languageService.\
                   getLanguageService(components.interfaces.koIAppInfoEx)
        rubyExe = rubyInfo.executablePath
        if not rubyExe:
            installations = rubyInfo.FindInstallationPaths()
            if installations:
                rubyExe = installations[0]
        if not rubyExe or not os.path.isfile(rubyExe):
            log.error("No Ruby interpreter could be found.")
            return ""
        return rubyExe
    
    def _getRubyIncArgv(self):
        argv = []
        rubyExtraPaths = []
        if self.prefService.prefs.hasPref("rubyExtraPaths"):
            rubyExtraPaths = self.prefService.prefs.getStringPref("rubyExtraPaths")
        if rubyExtraPaths:
            if sys.platform.startswith("win"):
                rubyExtraPaths = string.replace(rubyExtraPaths, '\\', '/')
            rubyExtraPaths = [x for x in rubyExtraPaths.split(os.pathsep) if x.strip()]
            for incpath in rubyExtraPaths:
                argv += ['-I', incpath]
        return argv

    def _getLogFile(self):
        # The RUBYDB_OPTS LogFile setting is given only if
        # logEnabled and len(logPath/logDir) > 0 (duh)
        # add logging if configured
        logEnabled = 0
        logDir = ''
        if self.prefService.prefs.hasPref("ruby_debuggerlogenabled"):
            logEnabled = self.prefService.prefs.getBooleanPref("ruby_debuggerlogenabled")
        if logEnabled:
            if self.prefService.prefs.hasPref("ruby_debuggerlogpath"):
                logDir = self.prefService.prefs.getStringPref("ruby_debuggerlogpath")
                logDir = os.path.expanduser(logDir)
            if not logDir:
                logDir = os.getenv("TEMP", os.getenv("TMP", ''))
                if logDir:
                    logDir = os.path.join(logDir, 'ruby-dbgp.log')
            if logDir:
                logDir = logDir.replace("\\", "/")
                # encode the path
                logDir = urllib.quote(logDir)
        return logDir
    
    def _getScriptArgs(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')
        # use the get variables as the user_args if emulating CGI        
        scriptArgs = []
        if self._usecgi:
            get_args = self._build_get_line()
            if get_args: scriptArgs.append(get_args)
        if len(scriptArgs) < 1 and \
           self.currentInstancePrefs.hasPref('params'):
            params = self.currentInstancePrefs.getStringPref('params')
            if params:
                scriptArgs = listcmd.line2argv(params)
        return scriptArgs
        
    def _getDebuggerEnvironment(self, installed=0):
        env = self._getenv()

        cmdPort = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        #listenerPort = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        rubydbOptString = 'RemotePort=127.0.0.1:%d' % (cmdPort) # Required
        
        logFile = self._getLogFile()
        if logFile:
            rubydbOptString = rubydbOptString + ' LogFile=' + logFile
        env["RUBYDB_OPTS"] = rubydbOptString
        if log.isEnabledFor(logging.DEBUG):
            log.debug("%s Ruby Debugging Environment", "="*40)
            log.debug("RUBYDB_OPTS=%s", env['RUBYDB_OPTS'])
            log.debug("="*70)
        return env

    def _fix_RubyDebuggerEnv(self, env):
        """ Stop the Ruby debugger from first stopping in ubygems.rb [sic]
        """
        if env.has_key('RUBYOPT') and env['RUBYOPT'].lower() == "rubygems":
            del env['RUBYOPT']

    def _launchForInteractive(self):
        executable = self.currentInstancePrefs.getStringPref('executable')
        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)
        argv = ([str(executable)] + execArgs +
                ["-I", self.dbgpLibDir,
                 "-r", self.dbgpLibDir + "/rdbgp.rb"] +
                self._getRubyIncArgv() +
                ["-e", "1"])
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.getcwd()
        env = self._getDebuggerEnvironment()
        self._fix_RubyDebuggerEnv(env)
        log.debug("About to launch the interactive shell: (%s, %s, %s)", argv, cwd, env)
        self._doLaunch(argv, cwd, env)

    def _launchForDebug(self):
        script = self.currentInstancePrefs.getStringPref('filename')
        executable = self.currentInstancePrefs.getStringPref('executable')
        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)
        argv = ([str(executable)] + execArgs +
                ["-I", self.dbgpLibDir,
                 "-r", self.dbgpLibDir + "/rdbgp.rb"] +
                self._getRubyIncArgv() +
                [script] +
                self._getScriptArgs())
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        env = self._getDebuggerEnvironment()
        # Uncomment this next line if you want to "fix" bug 43269
        # by always turning off automatically loading rubygems.
        # Currently (when bug 43269 was logged, 2005-12-07),
        # I was disabling the flag when doing the interactive shell,
        # but not for normal debugging.  I've decided to leave it as is,
        # because there currently is no pref to turn it on, and it will
        # get in the way of users who rely on that behavior.
        #
        # Uncomment this line to turn off the <-r ubygems.rb> flag when
        # debugging.  This will break code that uses third-party modules
        # and doesn't explicitly load rubygems.
        
        # self._fix_RubyDebuggerEnv(env)
        if env.has_key('RUBYOPT') and env['RUBYOPT'].lower() == "rubygems":
            log.debug("env['RUBYOPT'] = %s", env['RUBYOPT'])
        
        res = self._doLaunch(argv, cwd, env)
        return res

    def _launchForRun(self):
        script = self.currentInstancePrefs.getStringPref('filename')
        executable = self.currentInstancePrefs.getStringPref('executable')
        params = self.currentInstancePrefs.getStringPref('executable-params')
        execArgs = listcmd.line2argv(params)
        argv = ([str(executable)] + execArgs +
               self._getRubyIncArgv() + [script] +
               self._getScriptArgs())
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        env = self._getenv()
        self._doLaunch(argv, cwd, env)

# This class is currently unreachable, as all the UI code
# has been commented out.
class koJavaScriptInvocation(koInterpreterInvocation):
    _language = "JavaScript"
    _reg_clsid_ = "{03d6b8b6-c32d-4b47-bf9f-f6af939aecc6}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=JavaScript"
    #_reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "JavaScript"
    _invocation_description_ = "JavaScript"
    #XXX There are no JavaScript prefs, although we might want to give ability
    # to select a Firefox executable

    def get_supportedInvokeTypes(self):
        return INVOKE_INTERACTIVE

    def __init__(self):
        koInterpreterInvocation.__init__(self)
        # No need to create a prefs-observer as there aren't any JS prefs
        self.error_invocation_not_supported = "Komodo does not support %s for JavaScript directly, "+\
                                              "you must start from Firefox.  You can install the Firefox "+\
                                              "debugging extension from Preferences.  Please refer to help "+\
                                              "for more information."
        
    def _determineExecutable(self):
        koWebbrowser = components.classes['@activestate.com/koWebbrowser;1'] \
            .getService(components.interfaces.koIWebbrowser)
        firefoxes = koWebbrowser.get_firefox_paths()
        if len(firefoxes) == 0:
            log.error("No Firefox browser could be found.")
            return ""
        # Same as pref-javascript.js, return first if more than one
        #XXX Add prefered firefox to prefs
        return firefoxes[0]
    
    def _launchForInteractive(self):
        executable = self._determineExecutable()
        if not executable:
            log.debug("Can't find a JavaScript executable")
            return
        cmdPort = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        argv = ([str(executable),
                 '-jsdbgp',
                 'localhost:%d' % (cmdPort),
                 '-jsshell'])
        cwd = os.getcwd()
        env = self._getenv()
        log.debug("About to launch the interactive shell: (%s, %s, %s)", argv, cwd, env)
        self._doLaunch(argv, cwd, env)
        
    # Things expected by the environment:
    def _do_validate(self, invokeType):
        if invokeType == INVOKE_INTERACTIVE:
            return 1, None, None
        return 0, 42, "Can't debug JavaScript within Komodo; use Firefox"
        return 1, None, None
