#!python
# Copyright (c) 2004-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

"""The multilang-zone of the codeintel database.
See the database/database.py module docstring for an overview.
"""

import sys
import os
from os.path import (join, dirname, exists, expanduser, splitext, basename,
                     split, abspath, isabs, isdir, isfile)
from glob import glob
from pprint import pprint, pformat
import logging
from cStringIO import StringIO
import copy

import ciElementTree as ET
from codeintel2.common import *
from codeintel2.database.langlib import LangZone



#---- globals

log = logging.getLogger("codeintel.db")
#log.setLevel(logging.DEBUG)



#---- Database zone and lib implementations

class MultiLangTopLevelNameIndex(object):
    """A wrapper around the plain-dictionary toplevelname_index for a
    MultiLangZone dir to provide better performance for continual updating.

        {lang -> toplevelname -> blobnames}

    A 'toplevelname_index' is a merge of {lang -> blobname -> toplevelnames}
    data for all resources in its dir.  As those resources are
    continually re-scanned (e.g. as a file is edited in Komodo), it
    would be too expensive to update this index everytime.

    Solution:
    Keep a list of "recent updates" and only merge them into the main
    data when that buf hasn't been updated in "a while" and when needed
    for saving the index. Note: Buffer *removals* are not put on-deck,
    but removed immediately.

    # Notes on locking

    This class does not guard its datastructures with locking. It is up
    to the MultiLangZone using this to guard against simultaneous access
    on separate threads.
    """
    def __init__(self, data=None, timeout=90):
        # toplevelname_index data: {lang -> toplevelname -> blobnames}
        if data is None:
            self._data = {}
        else:
            self._data = data

        # Time (in seconds) to hold a change "on deck".
        # Timed-out changes are merged on .get() and .update().
        self.timeout = timeout
        self._on_deck = {
            # basename                           # the basename of the buf path
            #   -> [timestamp,                   # time of the last update
            #       {lang -> blobname -> toplevelnames}, # i.e. the dict in res_index
            #       {lang -> toplevelname -> blobnames}] # lazily generated pivot
        }

    def __repr__(self):
        return "<MultiLangTopLevelNameIndex: %d update(s) on-deck>"\
               % len(self._on_deck)

    def merge(self):
        """Merge all on-deck changes with `self.data'."""
        for base, (timestamp, toplevelnames_from_blobname_from_lang,
                blobnames_from_toplevelname_lang) in self._on_deck.items():
            if blobnames_from_toplevelname_lang is None:
                blobnames_from_toplevelname_lang = self._bftfl_from_tfbfl(
                    toplevelnames_from_blobname_from_lang)
            for lang, toplevelname, blobnames in (
                    (lang, toplevelname, blobnames)
                    for lang, bft in blobnames_from_toplevelname_lang.items()
                    for toplevelname, blobnames in bft.items()
                ):
                if toplevelname not in self._data:
                    self._data.setdefault(lang, {})[toplevelname] = blobnames
                else:
                    self._data.setdefault(lang, {})[toplevelname].update(blobnames)
            del self._on_deck[base]

    def merge_expired(self, now):
        """Merge expired on-deck changes with `self.data'."""
        for base, (timestamp, toplevelnames_from_blobname_from_lang,
                blobnames_from_toplevelname_lang) in self._on_deck.items():
            if now - timestamp < self.timeout:
                continue
            if blobnames_from_toplevelname_lang is None:
                blobnames_from_toplevelname_lang = self._bftfl_from_tfbfl(
                    toplevelnames_from_blobname_from_lang)
            for lang, toplevelname, blobnames in (
                    (lang, toplevelname, blobnames)
                    for lang, bft in blobnames_from_toplevelname_lang.items()
                    for toplevelname, blobnames in bft.items()
                ):
                if toplevelname not in self._data:
                    self._data.setdefault(lang, {})[toplevelname] = blobnames
                else:
                    self._data.setdefault(lang, {})[toplevelname].update(blobnames)
            del self._on_deck[base]

    @property
    def data(self):
        self.merge()
        return self._data

    def update(self, base, old_toplevelnames_from_blobname_from_lang,
               new_toplevelnames_from_blobname_from_lang):
        now = time.time()

        if base not in self._on_deck:
            # Remove old refs from current data.
            for lang, old_toplevelnames_from_blobname \
                    in old_toplevelnames_from_blobname_from_lang.iteritems():
                for blobname, toplevelnames \
                        in old_toplevelnames_from_blobname.iteritems():
                    for toplevelname in toplevelnames:
                        #TODO: catch KeyError here? (may indicate corruption)
                        self._data[lang][toplevelname].remove(blobname)
                        if not self._data[lang][toplevelname]:
                            del self._data[lang][toplevelname]
                if not self._data[lang]:
                    del self.data[lang]
        else:
            blobnames_from_toplevelname = None

        self._on_deck[base] = [now, new_toplevelnames_from_blobname_from_lang,
                               None]
        self.merge_expired(now)

    def remove(self, base, toplevelnames_from_blobname_from_lang):
        if base in self._on_deck:
            del self._on_deck[base]
        else:
            for lang, toplevelnames_from_blobname \
                    in toplevelnames_from_blobname_from_lang.iteritems():
                for blobname, toplevelnames in toplevelnames_from_blobname.iteritems():
                    for toplevelname in toplevelnames:
                        try:
                            self._data[lang][toplevelname].remove(blobname)
                        except KeyError, ex:
                            pass # ignore this for now, might indicate corruption
                        else:
                            if not self._data[lang][toplevelname]:
                                del self._data[lang][toplevelname]
                if not self._data[lang]:
                    del self._data[lang]

    def _bftfl_from_tfbfl(self, toplevelnames_from_blobname_from_lang):
        blobnames_from_toplevelname_from_lang = dict(
            (lang, {}) for lang in toplevelnames_from_blobname_from_lang
        )
        for lang, toplevelnames_from_blobname \
                in toplevelnames_from_blobname_from_lang.items():
            blobnames_from_toplevelname \
                = blobnames_from_toplevelname_from_lang[lang]
            for blobname, toplevelnames in toplevelnames_from_blobname.iteritems():
                for toplevelname in toplevelnames:
                    if toplevelname not in blobnames_from_toplevelname:
                        blobnames_from_toplevelname[toplevelname] = set([blobname])
                    else:
                        blobnames_from_toplevelname[toplevelname].add(blobname)
        return blobnames_from_toplevelname_from_lang

    def get(self, lang, toplevelname, default=None):
        self.merge_expired(time.time())
        # First check on-deck items.
        for base, (timestamp, toplevelnames_from_blobname_from_lang,
               blobnames_from_toplevelname_from_lang) in self._on_deck.items():
            if blobnames_from_toplevelname_from_lang is None:
                blobnames_from_toplevelname_from_lang \
                    = self._on_deck[base][2] \
                    = self._bftfl_from_tfbfl(toplevelnames_from_blobname_from_lang)
            try:
                return blobnames_from_toplevelname_from_lang[lang][toplevelname]
            except KeyError:
                pass
        # Then, fallback to already merged data.
        try:
            return self._data[lang][toplevelname]
        except KeyError:
            return default


class MultiLangZone(LangZone):
    toplevelname_index_class = MultiLangTopLevelNameIndex

    def get_lib(self, name, dirs, sublang):
        assert isinstance(dirs, (tuple, list))
        assert sublang is not None, "must specify '%s' sublang" % self.lang

        #TODO Better canonicalization of dirs.
        canon_dirs = tuple(isabs(d) and d or abspath(d) for d in dirs)
        key = (canon_dirs, sublang)
        if key in self._dirslib_cache:
            return self._dirslib_cache[key]

        langdirslib = MultiLangDirsLib(self, self._lock, self.lang,
                                        name, canon_dirs, sublang)
        
        N = 10
        while len(self._ordered_dirslib_cache_keys) >= N:
            cache_key = self._ordered_dirslib_cache_keys.pop()
            del self._dirslib_cache[cache_key]
        self._dirslib_cache[key] = langdirslib
        self._ordered_dirslib_cache_keys.insert(0, key)

        return langdirslib

    def dfb_from_dir(self, dir, sublang, default=None):
        """Get the {blobname -> dbfile} mapping index for the given dir
        and lang.
        
        'dfb' stands for 'dbfile_from_blobname'.
        This must be called with the lock held.
        """
        blob_index = self.load_index(dir, "blob_index", default=default)
        try:
            return blob_index[sublang]
        except KeyError, ex:
            if default is not None:
                return default
            raise

    def get_buf_data(self, buf):
        #TODO Canonicalize path (or assert that it is canonicalized)
        #     Should have a Resource object that we pass around that
        #     handles all of this.
        self._acquire_lock()
        try:
            dir, base = split(buf.path)
            res_index = self.load_index(dir, "res_index", {})
            if base not in res_index:
                raise NotFoundInDatabase("%s buffer '%s' not found in database"
                                         % (buf.lang, buf.path))
            scan_time, scan_error, toplevelnames_from_blobname_from_lang \
                = res_index[base]

            try:
                blob_index = self.load_index(dir, "blob_index")
            except EnvironmentError, ex:
                self.db.corruption("MultiLangZone.get_buf_data",
                    "could not find 'blob_index' index: %s" % ex,
                    "recover")
                raise NotFoundInDatabase("%s buffer '%s' not found in database"
                                         % (buf.lang, buf.path))

            dhash = self.dhash_from_dir(dir)
            blob_from_lang = {}
            for lang, blobname in (
                 (lang, tfb.keys()[0])
                 for lang, tfb in toplevelnames_from_blobname_from_lang.items()
                ):
                dbsubpath = join(dhash, blob_index[lang][blobname])
                try:
                    blob = self.load_blob(dbsubpath)
                except ET.XMLParserError, ex:
                    #XXX Or should we clean out index and raise NotFoundInDatabase?
                    self.db.corruption("MultiLangZone.get_buf_data",
                        "could not parse dbfile for '%s' blob: %s"\
                            % (blobname, ex),
                        "ignore")
                    continue
                except EnvironmentError, ex:
                    self.db.corruption("MultiLangZone.get_buf_data",
                        "could not read dbfile for '%s' blob: %s"\
                            % (blobname, ex),
                        "ignore")
                    continue
                assert blob.get("lang") == lang
                blob_from_lang[lang] = blob

            return scan_time, scan_error, blob_from_lang
        finally:
            self._release_lock()

    def remove_buf_data(self, buf):
        """Remove the given resource from the database."""
        #TODO Canonicalize path (or assert that it is canonicalized)
        #     Should have a Resource object that we pass around that
        #     handles all of this.
        self._acquire_lock()
        try:
            dir, base = split(buf.path)

            res_index = self.load_index(dir, "res_index", {})
            try:
                scan_time, scan_error, toplevelnames_from_blobname_from_lang \
                    = res_index[base]
            except KeyError:
                # This resource isn't loaded in the db. Nothing to remove.
                return

            try:
                blob_index = self.load_index(dir, "blob_index")
            except EnvironmentError, ex:
                self.db.corruption("MultiLangZone.remove_buf_data",
                    "could not read blob_index for '%s' dir: %s" % (dir, ex),
                    "recover")
                blob_index = {}

            is_hits_from_lpath_lang = self.lang in self.db.hits_from_lpath_langs
            if is_hits_from_lpath_lang:
                try:
                    toplevelname_index = self.load_index(dir, "toplevelname_index")
                except EnvironmentError, ex:
                    self.db.corruption("MultiLangZone.remove_buf_data",
                        "could not read toplevelname_index for '%s' dir: %s"
                            % (dir, ex),
                        "recover")
                    toplevelname_index = self.toplevelname_index_class()

            dhash = self.dhash_from_dir(dir)
            del res_index[base]
            for lang, blobname in (
                 (lang, tfb.keys()[0])
                 for lang, tfb in toplevelnames_from_blobname_from_lang.items()
                ):
                try:
                    dbfile = blob_index[lang][blobname]
                except KeyError:
                    blob_index_path = join(dhash, "blob_index")
                    self.db.corruption("MultiLangZone.remove_buf_data",
                        "%s '%s' blob not in '%s'" \
                            % (lang, blobname, blob_index_path),
                        "ignore")
                    continue
                del blob_index[lang][blobname]
                for path in glob(join(self.base_dir, dhash, dbfile+".*")):
                    log.debug("fs-write: remove %s|%s blob file '%s/%s'",
                              self.lang, lang, dhash, basename(path))
                    os.remove(path)
            if is_hits_from_lpath_lang:
                toplevelname_index.remove(base,
                    toplevelnames_from_blobname_from_lang)

            self.changed_index(dir, "res_index")
            self.changed_index(dir, "blob_index")
            if is_hits_from_lpath_lang:
                self.changed_index(dir, "toplevelname_index")
        finally:
            self._release_lock()
        #XXX Database.clean() should remove dirs that have no
        #    dbfile_from_blobname entries.

    def update_buf_data(self, buf, force=False):
        """Update this MultiLangZone with the buffer data.

            "buf" is the CitadelBuffer instance with the relevant info
                (path, tree, scan_time).
            "force" (default False) is a boolean indicating if the
                buffer data should be updated even if buf.scan_time is
                <= that in the database.
        """
        self._acquire_lock()
        try:
            #TODO: Canonicalize path (or assert that it is canonicalized)
            dir, base = split(buf.path)

            # Get the current data, if any.
            res_index = self.load_index(dir, "res_index", {})
            res_index_has_changed = False
            blob_index = self.load_index(dir, "blob_index", {})
            blob_index_has_changed = False
            is_hits_from_lpath_lang = self.lang in self.db.hits_from_lpath_langs
            if is_hits_from_lpath_lang:
                toplevelname_index = self.load_index(dir, "toplevelname_index", {})
                toplevelname_index_has_changed = False
            try:
                (old_scan_time, old_scan_error, 
                 old_toplevelnames_from_blobname_from_lang) = res_index[base]
            except KeyError:    # adding a new entry
                (old_scan_time, old_scan_error, 
                 old_toplevelnames_from_blobname_from_lang) = None, None, {}
            else:               # updating an existing entry
                if not force and buf.scan_time is not None \
                   and buf.scan_time <= old_scan_time:
                    log.debug("skipping db update for '%s': %s < %s and "
                              "no force option",
                              base, buf.scan_time, old_scan_time)
                    return

            log.debug("update from %s buf '%s'", buf.lang, buf.path)

            # Parse the tree and get the list of blobnames.
            new_toplevelnames_from_blobname_from_lang = {}
            new_blob_from_lang_and_blobname = {}
            for lang, blob in buf.blob_from_lang.items():
                blobname = blob.get("name")
                new_blob_from_lang_and_blobname[(lang, blobname)] = blob
                new_toplevelnames_from_blobname_from_lang[lang] \
                    = {blobname: blob.names.keys()}

            # Determine necessary changes to res_index.
            new_scan_time = buf.scan_time
            new_scan_error = buf.scan_error
            if new_scan_error:
                if (new_scan_time != old_scan_time
                    or new_scan_error != old_scan_error):
                    res_index[base] = (new_scan_time, new_scan_error,
                                old_toplevelnames_from_blobname_from_lang)
                    res_index_has_changed = True

            else:
                # Only consider new blobs if there wasn't a scan error.
                # I.e., we want to preserve the last good scan info.

                if (new_scan_time != old_scan_time
                    or new_scan_error != old_scan_error
                    or new_toplevelnames_from_blobname_from_lang
                       != old_toplevelnames_from_blobname_from_lang):
                    res_index[base] = (new_scan_time, new_scan_error,
                                new_toplevelnames_from_blobname_from_lang)
                    res_index_has_changed = True

                if is_hits_from_lpath_lang:
                    if (new_toplevelnames_from_blobname_from_lang
                        != old_toplevelnames_from_blobname_from_lang):
                        toplevelname_index.update(base,
                            old_toplevelnames_from_blobname_from_lang,
                            new_toplevelnames_from_blobname_from_lang)
                        toplevelname_index_has_changed = True

                # Determine necessary changes to dbfile_from_blobname index
                # and the dbfiles and then make them.
                dbfile_changes = []
                for (lang, blobname), blob \
                        in new_blob_from_lang_and_blobname.items():
                    try:
                        old_toplevelnames_from_blobname_from_lang[lang][blobname]
                    except KeyError:
                        dbfile_changes.append(("add", lang, blobname, blob))
                    else:
                        dbfile_changes.append(("update", lang, blobname, blob))
                for lang, old_toplevelnames_from_blobname \
                        in old_toplevelnames_from_blobname_from_lang.items():
                    for blobname in old_toplevelnames_from_blobname:
                        try:
                            new_toplevelnames_from_blobname_from_lang[lang][blobname]
                        except KeyError:
                            dbfile_changes.append(("remove", lang, blobname, None))

                dhash = self.dhash_from_dir(dir)
                for action, lang, blobname, blob in dbfile_changes:
                    if action == "add":
                        dbfile = self.db.bhash_from_blob_info(
                                    buf.path, lang, blobname)
                        blob_index.setdefault(lang, {})[blobname] = dbfile
                        blob_index_has_changed = True
                        dbdir = join(self.base_dir, dhash)
                        if not exists(dbdir):
                            self._mk_dbdir(dbdir, dir)
                        #XXX What to do on write failure?
                        log.debug("fs-write: %s|%s blob '%s/%s'",
                                  self.lang, lang, dhash, dbfile)
                        if blob.get("src") is None:
                            blob.set("src", buf.path)   # for defns_from_pos() support
                        ET.ElementTree(blob).write(join(dbdir, dbfile+".blob"))
                    elif action == "remove":
                        dbfile = blob_index[lang][blobname]
                        del blob_index[lang][blobname]
                        blob_index_has_changed = True
                        #XXX What to do on removal failure?
                        log.debug("fs-write: remove %s|%s blob '%s/%s'",
                                  self.lang, lang, dhash, dbfile)
                        try:
                            os.remove(join(self.base_dir, dhash, dbfile+".blob"))
                        except EnvironmentError, ex:
                            self.db.corruption("MultiLangZone.update_buf_data",
                                "could not remove dbfile for '%s' blob: %s"\
                                    % (blobname, ex),
                                "ignore")
                    elif action == "update":
                        # Try to only change the dbfile on disk if it is
                        # different.
                        s = StringIO()
                        if blob.get("src") is None:
                            blob.set("src", buf.path)   # for defns_from_pos() support
                        ET.ElementTree(blob).write(s)
                        new_dbfile_content = s.getvalue()
                        dbfile = blob_index[lang][blobname]
                        dbpath = join(self.base_dir, dhash, dbfile+".blob")
                        # PERF: Might be nice to cache the new dbfile
                        #       content for the next time this resource is
                        #       updated. For files under edit this will be
                        #       common. I.e. just for the "editset".
                        try:
                            fin = open(dbpath, 'r')
                        except OSError, ex:
                            # Technically if the dbfile doesn't exist, this
                            # is a sign of database corruption. No matter
                            # though (for this blob anyway), we are about to
                            # replace it.
                            old_dbfile_content = None
                        else:
                            try:
                                old_dbfile_content = fin.read()
                            finally:
                                fin.close()
                        if new_dbfile_content != old_dbfile_content:
                            if not exists(dirname(dbpath)):
                                self._mk_dbdir(dirname(dbpath), dir)
                            #XXX What to do if fail to write out file?
                            log.debug("fs-write: %s|%s blob '%s/%s'",
                                      self.lang, lang, dhash, dbfile)
                            fout = open(dbpath, 'w')
                            try:
                                fout.write(new_dbfile_content)
                            finally:
                                fout.close()

            if res_index_has_changed:
                self.changed_index(dir, "res_index")
            if blob_index_has_changed:
                self.changed_index(dir, "blob_index")
            if is_hits_from_lpath_lang and toplevelname_index_has_changed:
                self.changed_index(dir, "toplevelname_index")
        finally:
            self._release_lock()
        #TODO: Database.clean() should remove dirs that have no
        #      blob_index entries.    



class MultiLangDirsLib(object):
    """A zone providing a view into an ordered list of dirs in a
    db/$multilang/... area of the db.

    These are dished out via Database.get_lang_lib(), which indirectly
    then is dished out by the MultiLangZone.get_lib(). Mostly this is
    just a view on the MultiLangZone singleton for this particular
    language.
    """
    def __init__(self, lang_zone, lock, lang, name, dirs, sublang):
        self.lang_zone = lang_zone
        self._lock = lock
        self.mgr = lang_zone.mgr
        self.lang = lang
        self.name = name
        self.dirs = dirs
        self.sublang = sublang
        self.import_handler \
            = self.mgr.citadel.import_handler_from_lang(sublang)

        # We keep a "weak" merged cache of blobname lookup for all dirs
        # in this zone -- where "weak" means that we verify a hit by
        # checking the current real dbfile_from_blobname index for that
        # dir (which may have changed). This caching slows down lookup
        # for single-dir LangDirsZones, but should scale better for
        # LangDirsZones with many dirs. (TODO-PERF: test this assertion.)
        self._dir_and_blobbase_from_blobname = {}
        self._importables_from_dir_cache = {}

    def __repr__(self):
        return "<%s %s>" % (self.lang, self.name)

    def _acquire_lock(self):
        self._lock.acquire()
    def _release_lock(self):
        self._lock.release()

    def has_blob(self, blobname, ctlr=None):
        dbsubpath = self._dbsubpath_from_blobname(blobname, ctlr=ctlr)
        return dbsubpath is not None

    def has_blob_in_db(self, blobname, ctlr=None):
        """Return true if the blobname is in the database.

        Typically this method is only used for debugging and .has_blob()
        is what you want.
        """
        dbsubpath = self._dbsubpath_from_blobname(
            blobname, ctlr=ctlr, only_look_in_db=True)
        return dbsubpath is not None

    def get_blob(self, blobname, ctlr=None):
        self._acquire_lock()
        try:
            dbsubpath = self._dbsubpath_from_blobname(
                blobname, ctlr=ctlr)
            if dbsubpath is not None:
                return self.lang_zone.load_blob(dbsubpath)
            else:
                return None
        finally:
            self._release_lock()

    def _importables_from_dir(self, dir):
        if dir not in self._importables_from_dir_cache:
            self._importables_from_dir_cache[dir] \
                = self.import_handler.find_importables_in_dir(dir)
        return self._importables_from_dir_cache[dir]
    
    def _dbsubpath_from_blobname(self, blobname, ctlr=None, 
                                 only_look_in_db=False):
        """Return the subpath to the dbfile for the given blobname,
        or None if not found.

        Remember that this is complicated by possible multi-level
        imports. E.g. "include('foo/bar.php')".
        """
        lang_zone = self.lang_zone

        self._acquire_lock()
        try:
            # Use our weak cache to try to return quickly.
            if blobname in self._dir_and_blobbase_from_blobname:
                blobdir, blobbase = self._dir_and_blobbase_from_blobname[blobname]

                # Check it. The actual info for that dir may have changed.
                dbfile_from_blobname = lang_zone.dfb_from_dir(blobdir, self.sublang)
                if blobbase in dbfile_from_blobname:
                    log.debug("have blob '%s' in '%s'? yes (in weak cache)",
                              blobname, blobdir)
                    return join(lang_zone.dhash_from_dir(blobdir),
                                dbfile_from_blobname[blobbase])
                del self._dir_and_blobbase_from_blobname[blobname] # drop from weak cache

            # Brute force: look in each dir.
            assert self.import_handler.sep is not None, \
                "%r.sep is None, this must be set" % self.import_handler
            blobparts = blobname.split(self.import_handler.sep)
            blobbase = blobparts[-1]
            for dir in self.dirs:
                if ctlr and ctlr.is_aborted():
                    log.debug("aborting search for blob '%s' on %s: ctlr aborted",
                              blobname, self)
                    return None

                # Is the blob in 'blobdir' (i.e. a non-multi-level import
                # that has been scanned already).
                blobdir = join(dir, *blobparts[:-1])
                dbfile_from_blobname = lang_zone.dfb_from_dir(
                                            blobdir, self.sublang, {})
                if blobbase in dbfile_from_blobname:
                    self._dir_and_blobbase_from_blobname[blobname] \
                        = (blobdir, blobbase)
                    log.debug("have blob '%s' in '%s'? yes (in dir index)", 
                              blobname, blobdir)
                    return join(lang_zone.dhash_from_dir(blobdir),
                                dbfile_from_blobname[blobbase])

                importables = self._importables_from_dir(blobdir)
                # 'importables' look like, for PHP:
                #   {'foo.php': ('foo.php', None, False),
                #    'foo.inc': ('foo.inc', None, False),
                #    'somedir': (None,      None, True)}

                if blobbase not in importables:
                    continue

                blobfile, subdir_blobbase, is_dir_import = importables[blobbase]
                if blobfile is None:
                    # There isn't an actual importable file here -- just
                    # a dir prefix to a multidir import.
                    log.debug("have blob '%s' in %s? no", blobname, self)
                    return None
                elif os.sep in blobfile:
                    # This is an import from a subdir. We need to get a new dbf.
                    blobdir = join(blobdir, dirname(blobfile))
                    blobfile = basename(blobfile)
                    blobbase = subdir_blobbase
                    dbfile_from_blobname = lang_zone.dfb_from_dir(
                                                blobdir, self.sublang, {})
                    if blobbase in dbfile_from_blobname:
                        self._dir_and_blobbase_from_blobname[blobname] \
                            = (blobdir, blobbase)
                        log.debug("have blob '%s' in '%s'? yes (in dir index)", 
                                  blobname, blobdir)
                        return join(lang_zone.dhash_from_dir(blobdir),
                                    dbfile_from_blobname[blobbase])

                # The file isn't loaded.
                if not only_look_in_db:
                    log.debug("%s importables in '%s':\n    %s", self.sublang,
                              blobdir, importables)
                    log.debug("'%s' likely provided by '%s' in '%s': "
                              "attempting load", blobname, blobfile, blobdir)
                    buf = self.mgr.buf_from_path(
                            join(blobdir, blobfile), self.lang)
                    buf.load()

                    dbfile_from_blobname = lang_zone.dfb_from_dir(
                                                blobdir, self.sublang, {})
                    if blobbase in dbfile_from_blobname:
                        self._dir_and_blobbase_from_blobname[blobname] \
                            = (blobdir, blobbase)
                        log.debug("have blob '%s' in '%s'? yes (after load)",
                                  blobname, blobdir)
                        return join(lang_zone.dhash_from_dir(blobdir),
                                    dbfile_from_blobname[blobbase])

            log.debug("have blob '%s' in %s? no", blobname, self)
            return None
        finally:
            self._release_lock()


