from xpcom import components, ServerException
import timeline

from koLanguageServiceBase import *
from koUDLLanguageBase import KoUDLLanguage

log = logging.getLogger('koXMLLanguageBase')

class koXMLLanguageBase(KoUDLLanguage):

    primary = 1
    styleBits = 6
    supportsBraceHighlighting = 0
    commentDelimiterInfo = { "block": [ ("<!--", "-->") ]  }
    _indent_chars = u'<>'
    _indent_open_chars = u'>'
    _indent_close_chars = u'>'
    _lineup_chars = '' # don't indent ()'s and the like in XML/HTML!
    _lineup_open_chars = "" # don't indent ()'s and the like in XML/HTML!

    supportsSmartIndent = "XML"

    def get_linter(self):
        if self._linter is None:
            self._linter = KoLinterLanguageService("@activestate.com/koXMLLinter;1")
        return self._linter

    def getEncodingWarning(self, encoding):
        if not encoding.use_byte_order_marker:
            if encoding.python_encoding_name.startswith('utf-16') or encoding.python_encoding_name.startswith('ucs-'):
                return 'Including a signature (BOM) is recommended for "%s".' % encoding.friendly_encoding_name
            else:
                return ''
        else: # It's all good
            return ''

    def _guessIndentation(self, scimoz, tabWidth, style_info):
        opener, indented = _findIndent(scimoz,
                                       self.stylingBitsMask,
                                       self._indent_open_chars,
                                       style_info._indent_open_styles,
                                       style_info._comment_styles,
                                       style_info._indent_open_opening_styles)
        if opener and indented:
            log.info("opener: %r" % opener)
            log.info("indented: %r" % indented)
            raw, indentsmall, foundTabsInOpener = classifyws(opener, tabWidth)
            raw, indentlarge, foundTabsInIndented = classifyws(indented, tabWidth)
            foundTabs = foundTabsInOpener or foundTabsInIndented
        else:
            indentsmall = indentlarge = 0
            foundTabs = 0
        guess = indentlarge - indentsmall
        if guess > 1:
            # indents less than 2 aren't considered reasonable guesses
            return guess, foundTabs
        return 0,0

class koHTMLLanguageBase(koXMLLanguageBase):

    def get_linter(self):
        if self._linter is None:
            self._linter = KoLinterLanguageService("@activestate.com/koHTMLLinter;1")
        return self._linter


def _findIndent(scimoz, bitmask, chars, styles, comment_styles, opening_styles):
    indenting = None
    timeline.enter("koXMLLanguage:_findIndent")
    try:
        # first, colourise the first 100 lines at most
        N = min(100, scimoz.lineCount-1)
        end = scimoz.getLineEndPosition(N)
        if scimoz.endStyled < end:
            scimoz.colourise(scimoz.endStyled, end)
        data = scimoz.getStyledText(0, end)
        # data is a list of (character, styleNo)
        
        for lineNo in range(N):
            if not scimoz.getLineIndentation(lineNo): # skip unindented lines
                continue
            lineEndPos = scimoz.getLineEndPosition(lineNo)
            lineStartPos = scimoz.positionFromLine(lineNo)
            line = scimoz.getTextRange(lineStartPos, lineEndPos)
            #start = lineEndPos
            # we're looking for the 'indenting' line
            for pos in range(lineEndPos, lineStartPos-1, -1):
                char = data[pos*2]
                style = ord(data[pos*2+1]) & bitmask
                if (char in chars) and (style in styles):
                    # partial success - we found that the first 'interesting'
                    # character from the right of the line is an
                    # indent-causing character
    
                    # now we need to find the line that _starts_ the XML tag.
                    # First look back in the document until you find a "<" character
                    # styled properly
                    for pos in range(pos, 0, -1):
                        char = data[pos*2]
                        style = ord(data[pos*2+1])
                        if char in ('<', u'<') and style in opening_styles:
                            startLineNo = scimoz.lineFromPosition(pos)
                            lineEndPos = scimoz.getLineEndPosition(startLineNo)
                            lineStartPos = scimoz.positionFromLine(startLineNo)
                            line = scimoz.getTextRange(lineStartPos, lineEndPos)
                            indenting = line
                            log.info("Found indenting line: %r" % line)
                            break
                    else:
                        log.info("couldn't find < tag")
                    break
                if char in string.whitespace:
                    # skip whitespace
                    continue
                if style in comment_styles:
                    # skip comments
                    continue
            if indenting: break
        else:
            log.info("Couldn't find an indenting line")
            return '', ''
        lineNo += 1
        for lineNo in range(lineNo, N):
            lineEndPos = scimoz.getLineEndPosition(lineNo)
            lineStartPos = scimoz.positionFromLine(lineNo)
            line = scimoz.getTextRange(lineStartPos, lineEndPos)
            # we want to skip lines that are just comments or just whitespace
            for pos in range(lineEndPos, lineStartPos-1, -1):
                char = scimoz.getWCharAt(pos)
                style = scimoz.getStyleAt(pos) & bitmask
                if char in string.whitespace:
                    # skip whitespace
                    continue
                if style in comment_styles:
                    # skip comments
                    continue
                # if we get here it must be a 'useful' line.
                indented = line
                log.info("Found indented line: %r" % line)
                return indenting, indented
        else:
            log.info("Couldn't find an indented line")
            return '', ''
    finally:
        timeline.leave("koXMLLanguage:_findIndent")

# taken from IDLE
# Look at the leading whitespace in s.
# Return pair (# of leading ws characters,
#              effective # of leading blanks after expanding
#              tabs to width tabwidth)

def classifyws(s, tabwidth):
    foundTabs = 0
    raw = effective = 0
    for ch in s:
        if ch == ' ':
            raw = raw + 1
            effective = effective + 1
        elif ch == '\t':
            foundTabs = 1
            raw = raw + 1
            effective = (effective / tabwidth + 1) * tabwidth
        else:
            break
    return raw, effective, foundTabs

