/* Copyright (c) 2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Source Code Control history dialog.
 *
 * To use this dialog you should just call sccdialogs.js::dialog_history().
 *
 * Usage:
 *  All dialog interaction is done via an object passed in and out as the first
 *  window argument: window.arguments[0].
 *      .urls           is the urls that have been selected by the user.
 *                      The URLs are automatically converted for display via
 *                      uriparse_displayPath().
 *      .localRevisions is the revisions numbers for the given urls. The length
 *                      of the localRevisions array must match the length of
 *                      urls.
 *
 * How it works:
 *  Once the dialog is given the list of urls and has been opened, it will
 *  display the list of urls in a selectable tree. Once the user selects a
 *  url (can also happen on load), the history for that url will then be shown
 *  in the history tree. The history items can be selected and a diff can then
 *  be made on a particular revision or between two revisions.
 *  
 * Dev Notes:
 *  This dialog is similar in implementation to commit.xul (xbl me!).
 */

var gSCCHistoryDialog_log = getLoggingMgr().getLogger("scc.history");
//gSCCHistoryDialog_log.setLevel(LOG_DEBUG);

var gSCCHistoryDialog = null;

function _sccHistoryTreeView(tree_type, initial_rows) {
    /* tree_type: 0 - urls, 1 - history items */
    this._history_tree_type = tree_type;
    this._rows = initial_rows;
    /* localRevision is used to show which is the current version the user has */
    this._localRevisionHistoryItem = null;
    this._debug = 0;
}

_sccHistoryTreeView.prototype = {
  // The nsITreeView object for the tree view
    getCellText : function(row, column) {
        if (this._history_tree_type == 0) {
            return uriparse_displayPath(this._rows[row]);
        } else {
            var koHistoryItem = this._rows[row];
            switch (column.id) {
                case 'scc_history_items_tree_column_revision':
                    // Mark the current revision row with a " (your copy)" text
                    if (koHistoryItem == this._localRevisionHistoryItem) {
                        return koHistoryItem.version + " (local revision)";
                    }
                    return koHistoryItem.version;
                case 'scc_history_items_tree_column_change':
                    return koHistoryItem.change;
                case 'scc_history_items_tree_column_user':
                    return koHistoryItem.author;
                case 'scc_history_items_tree_column_date':
                    return koHistoryItem.date;
                case 'scc_history_items_tree_column_action':
                    return koHistoryItem.action;
                case 'scc_history_items_tree_column_message':
                    return koHistoryItem.message;
            }
            return "(Unknown column: " + column.id + ")";
        }
    },
    get rowCount() { return this._rows.length; },

    // Everything below here is just to satisfy the interface, and not all of it may
    // be required.
    tree : null,
    isSeparator : function(index) {return false;},
    isContainer : function(index) {return false;},
    setTree : function(out) { this.tree = out; },
    getRowProperties : function(row,prop) {},
    getColumnProperties : function(column,prop){},
    getCellProperties : function(row,column,properties) {},
    isContainer : function(row) {return false;},
    cycleCell: function(row, colId) {},
    selectionChanged : function() {},
    performAction : function(action) {},
    isSorted : function() {return false;},
    getImageSrc : function() {return null;},
    cycleHeader : function() {},

    /* our own defined methods */
    /* XXX - Sorts correctly, but updating the rows re-selecting the row is failing */
    _sortColumn : null,
    _sortDirection : 0,  /* 0 - descending, 1 - ascending */

    //
    // XXX - Could optimize this function somewhat:
    //       By caching the sorted info, so no resorting needed when switching
    //       the sort order between columns.
    //
    sortByColumn : function(col, sortDirection)
    {
        // Function to compare two revisions
        // revision1 and revision2 are an array of ints, ex: [1, 565, 2, 22]
        function _historyCompare(compare1, compare2)
        {
            if (compare1 && compare2) {
                var value1 = compare1.rowValue;
                var value2 = compare2.rowValue;
                if (compare1.useArraySort) {
                    for (var i=0; i < value1.length; i++) {
                        if (i >= value2.length) {
                            //dump("Comparing " + value1 + " and " + value2 + "\n");
                            return -1;
                        } else if (value1[i] > value2[i]) {
                            return -1;
                        } else if (value1[i] < value2[i]) {
                            return 1;
                        }
                        /* else, they are the same and we continue looping */
                    }
                } else {
                    if (value1 > value2) {
                        return -1;
                    }
                }
            }
            return 1;
        };

        //dump("Sorting direction: " + this._sortDirection + "\n");
        //dump("Sorting by column: " + col.id + "\n");
        /* if it's the same as last sorted column, just reverse the order */
        if (this._sortColumn == col) {
            //dump("Just reversing sort order for existing column.\n");
            var newCurrentIndex = (this._rows.length - this.selection.currentIndex) - 1;
            this._rows.reverse();
            this.setTreeRows(this._rows);
            this.selection.currentIndex = newCurrentIndex;
            this.tree.ensureRowIsVisible(newCurrentIndex);
            this.setSortDirection(col, !this._sortDirection);
            return;
        } else if (!sortDirection || typeof(sortDirection) == 'undefined') {
            sortDirection = 1; /* ascending then */
            if (col.id == "scc_history_items_tree_column_revision" ||
                col.id == "scc_history_items_tree_column_change" ||
                col.id == "scc_history_items_tree_column_date") {
                sortDirection = 1; /* ascending then */
            } else {
                sortDirection = 0; /* descending then */
            }
        }

        // Get all the row data we need to compare with
        var unsorted_row_values = [];
        //dump("Getting row values...\n");
        if (col.id == "scc_history_items_tree_column_revision") {
            // Convert all fields to an array of ints
            var split;
            for (var i=0; i < this._rows.length; i++) {
                split = this._rows[i].version.split(".");
                for (var j=0; j < split.length; j++) {
                    split[j] = parseInt(split[j]);
                }
                unsorted_row_values[i] = { "useArraySort": true,
                                           "rowValue": split,
                                           "koHistoryItem": this._rows[i] };
            }
        } else if (col.id == "scc_history_items_tree_column_change") {
            // Convert all fields to ints
            for (var i=0; i < this._rows.length; i++) {
                unsorted_row_values[i] = { "useArraySort": false,
                                           "rowValue": parseInt(this.getCellText(i, col)),
                                           "koHistoryItem": this._rows[i] };
            }
        } else {
            // Convert all fields to a string value
            for (var i=0; i < this._rows.length; i++) {
                unsorted_row_values[i] = { "useArraySort": false,
                                           "rowValue": this.getCellText(i, col),
                                           "koHistoryItem": this._rows[i] };
            }
        }

        //dump("Beginning sort of " + unsorted_row_values.length + " rows...\n");
        var newCurrentIndex = -1;
        var selectedRevision = "";
        if (this.selection.currentIndex >= 0) {
            selectedRevision = this._rows[this.selection.currentIndex].version;
        }

        // Sort the rows, using our own customized sorting function
        unsorted_row_values.sort(_historyCompare);
        if (sortDirection == 0) {
            // Reverse the sorted order
            unsorted_row_values.reverse();
        }

        // Dump sorted rows back into self._rows
        var sorted_rows = [];
        for (var i=0; i < unsorted_row_values.length; i++) {
            sorted_rows.push(unsorted_row_values[i].koHistoryItem);
            // Keep track of the current row selected
            if ((newCurrentIndex == -1) &&
                sorted_rows[i].version == selectedRevision) {
                newCurrentIndex = i;
            }
        }
        this.setTreeRows(sorted_rows);

        this.selection.currentIndex = newCurrentIndex;
        this.tree.ensureRowIsVisible(newCurrentIndex);
        this.setSortDirection(col, sortDirection);
    },
    setSortDirection: function(column, newSortDirection) {
        // Remove old sort direction
        if (this._sortColumn && this._sortColumn != column) {
            this._sortColumn.removeAttribute("sortDirection");
        }

        // Update the sort direction on the sorted column
        if (newSortDirection == 0) {
            column.setAttribute("sortDirection", "ascending");
        } else {
            column.setAttribute("sortDirection", "descending");
        }
        this._sortColumn = column;
        this._sortDirection = newSortDirection;
    },
    sortAscending: function(column) {
        this.sortByColumn(column, 0);
    },
    sortDescending: function(column) {
        this.sortByColumn(column, 1);
    },
    markRevisionAsLocal: function(revision) {
        this._localRevisionHistoryItem = null;
        // Find and select the row with this revision
        for (var i=0; i < this._rows.length; i++) {
            if (this._rows[i].version == revision) {
                this.selection.select(i);
                this._localRevisionHistoryItem = this._rows[i];
                break;
            }
        }
    },
    /* Own function for updating the tree rows */
    setTreeRows : function(rows) {
        // Clearing the selection
        this.selection.clearSelection();
        // Using rowCountChanged to notify rows were removed
        this.tree.rowCountChanged(0, -this._rows.length);
        this._rows = rows;
        // Using rowCountChanged to notify rows were added
        this.tree.rowCountChanged(0, this._rows.length);
    }
};


//---- interface routines for XUL

function scc_history_OnLoad()
{
    try {
        document.title = "SCC History";

        // .urls
        var urls = window.arguments[0].urls;
        if (typeof urls == "undefined" || urls == null || urls.length < 1) {
            //XXX Is this the kind of error handling we want to do in onload
            //    handlers?
            var msg = "Internal Error: illegal 'urls' value for "
                      +"SCC History dialog: '"+urls+"'.";
            gSCCHistoryDialog_log.error(msg);
            alert(msg);
            window.close();
        }

        var localRevisions = window.arguments[0].localRevisions;
        if (typeof localRevisions == "undefined" || localRevisions == null ||
            localRevisions.length != urls.length) {
            var msg = "Internal Error: illegal 'localRevisions' value for "
                      +"SCC History dialog: '"+localRevisions+"'.";
            gSCCHistoryDialog_log.error(msg);
            alert(msg);
            window.close();
        }

        // .sccType
        var scc_type = window.arguments[0].sccType;
        // .sccHandler
        var scc_handler = window.arguments[0].sccHandler;
        if (typeof scc_type == "undefined" || scc_type == null ||
            typeof scc_handler == "undefined" || scc_handler == null) {
            //XXX Is this the kind of error handling we want to do in onload
            //    handlers?
            var msg = "Internal Error: illegal 'sccHandler' or 'sccType' value for "
                      +"SCC History dialog: '"+scc_handler+"', '"+scc_type+"'.";
            gSCCHistoryDialog_log.error(msg);
            alert(msg);
            window.close();
        }
    
        gSCCHistoryDialog = new _scc_history_dialog(scc_type, scc_handler, urls,
                                                    localRevisions);

        // Select the first url in the given list
        if (urls && urls.length > 0) {
            gSCCHistoryDialog.urlsTreeView.selection.select(0);
        }

        // Listen for keypress events (on capture)
        window.addEventListener("keypress", _scc_history_keypressHandler, true);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
        alert(e);
        window.close();
    }
}

function scc_history_OnUnload()
{
    window.removeEventListener("keypress", _scc_history_keypressHandler, true);
}


function _scc_history_keypressHandler(event)
{
    gSCCHistoryDialog.handleKeypress(event);
}


function _scc_history_dialog(sccType, sccHandler, urls, localRevisions)
{
    try {
        // String: "p4", "cvs" or "svn"
        this.sccType = sccType;
        // koISCC xpcom object
        this.sccHandler = sccHandler;
        // Array of strings
        this.urls = urls;
        // Array of strings, one for each url in urls
        this.localRevisions = localRevisions;
        // The local revision for the url currently selected
        this.localRevisionForUrl = "";
        // Last error service
        this._lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                                .getService(Components.interfaces.koILastErrorService);
        // Prefs
        this._prefsSvc = Components.classes["@activestate.com/koPrefService;1"].
                                getService(Components.interfaces.koIPrefService);
        this._prefsSvc = this._prefsSvc.prefs;

        // Find some xul elements
        this.dialog = document.getElementById("dialog-scc-history");
        this.messageTextbox = document.getElementById("scc_history_message");
        this.urlsVbox = document.getElementById("scc_history_urls_vbox");
        this.urlsTree = document.getElementById("scc_history_urls_tree");
        this.historyTree = document.getElementById("scc_history_items_tree");
        this.history_revision_column = document.getElementById("scc_history_items_tree_column_revision");
        this.history_change_column = document.getElementById("scc_history_items_tree_column_change");
        this.history_action_column = document.getElementById("scc_history_items_tree_column_action");
        this.okButton = this.dialog.getButton("accept");
        this.cancelButton = this.dialog.getButton("cancel");
        this.cmd_diffToLocalVersion   = document.getElementById("cmd_sccHistory_diffToLocalVersion");
        this.cmd_diffToCurrentVersion = document.getElementById("cmd_sccHistory_diffToCurrentVersion");
        this.cmd_diffBetweenVersions  = document.getElementById("cmd_sccHistory_diffBetweenVersions");
        // Setup button access keys
        this.okButton.setAttribute("accesskey", "o");
        this.cancelButton.setAttribute("accesskey", "c");

        // show url section tree if there is more than one url
        if (this.urls.length > 1) {
            this.urlsVbox.setAttribute("collapsed", false);
            // get taller if there are a lot of urls
            if (this.urls.length > 20) {
                addStyle(this.urlsTree, "min-height", "400px");
            }
        } else {
            this.urlsVbox.setAttribute("collapsed", true);
        }

        // Action is only applicable for perforce
        if (this.sccType != "p4") {
            this.history_action_column.setAttribute("hidden", "true");
            this.history_change_column.setAttribute("hidden", "true");
        }

        // Setup the tree views
        this.urlsTreeView = new _sccHistoryTreeView(0, this.urls);
        this.urlsTree.treeBoxObject.view = this.urlsTreeView;

        this.historyTreeView = new _sccHistoryTreeView(1, []);
        this.historyTree.treeBoxObject.view = this.historyTreeView;

        // Size to content before moving so "moveToAlertPosition" calculations are
        // correct.
        window.sizeToContent();
        this.dialog.moveToAlertPosition();
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

// Load history info for a url
_scc_history_dialog.prototype.load_history_for_url = function (url)
{
    try {
        var options = "";
        switch (this.sccType) {
            case 'p4':
                // XXX - Move these to a user controlled widget
                options = "L";    /* longFormat */
                options += "i";   /* follow branches */
                break;
        }
        // history will be an array of koISCCHistoryItem
        var history = this.sccHandler.history(url,  /* scc file   */
                                              options,
                                              new Object() /* array length */);

        //dump("history.length: " + history.length + "\n");
        var error = this._lastErrorSvc.getLastErrorMessage();
        if (error) {
            alert(error);
            return;
        }

        // process history now and throw it into the tree
        this.historyTreeView.setTreeRows(history);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

// Load history items for the given url
_scc_history_dialog.prototype.urls_onSelectionChange = function ()
{
    try {
        // Update the history items tree
        var uri = this.urls[this.urlsTree.currentIndex];
        this.localRevisionForUrl = this.localRevisions[this.urlsTree.currentIndex];
        this.load_history_for_url(uri);
        // Select the first history item (most recent) in the list
        if (this.historyTreeView.rowCount > 0) {
            // Initially, sort the columns in descending revision order
            // XXX - For perforce, we can list branch info, which restarts the
            //       revision number for that file, so perhaps we should sort
            //       by change number for perforce
            if (this.sccType == "p4") {
                this.historyTreeView.sortDescending(this.history_change_column);
            } else {
                this.historyTreeView.sortDescending(this.history_revision_column);
            }
            this.historyTreeView.markRevisionAsLocal(this.localRevisionForUrl);
            //this.historyTreeView.selection.select(0);
        }
        // Title: "filename : revision xyz"
        document.title = uriparse_baseName(uri) + " : local revision is " + this.localRevisionForUrl;
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

// Load history details for the given history item
_scc_history_dialog.prototype.history_onSelectionChange = function ()
{
    try {
        //dump("history_onSelectionChange\n");
        // Update the history message box
        if (this.historyTreeView._rows && this.historyTree.currentIndex >= 0) {
            //dump("this.historyTreeView._rows: "  + this.historyTreeView._rows + "\n");
            //dump("this.historyTree.currentIndex: "  + this.historyTree.currentIndex + "\n");
            var message = this.historyTreeView._rows[this.historyTree.currentIndex].message;
            this.messageTextbox.value = message;
        }

        this.cmd_diffToLocalVersion.setAttribute("disabled",
                                                   (this.historyTreeView.selection.count != 1));
        this.cmd_diffToCurrentVersion.setAttribute("disabled",
                                                   (this.historyTreeView.selection.count != 1));
        this.cmd_diffBetweenVersions.setAttribute("disabled",
                                                   (this.historyTreeView.selection.count != 2));
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

_scc_history_dialog.prototype.sortHistoryTreeByColumn = function (treecol)
{
    try {
        this.historyTreeView.sortByColumn(treecol);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

_scc_history_dialog.prototype.performDiff = function (uri1, version1,
                                                      uri2, version2,
                                                      filepath, diffToLocal)
{
    try {
        // Get diff options
        var options = '';
        var external = '';
        switch (this.sccType) {
            case 'p4':
                if (this._prefsSvc.getStringPref('p4_diff_type') == 'komododiff') {
                    options = this._prefsSvc.getStringPref('p4DiffOptions');
                } else {
                    options = '';
                    external = this._prefsSvc.getStringPref('externaldiff');
                }
                if (diffToLocal) {
                    if (this.historyTreeView._localRevisionHistoryItem &&
                        (this.historyTreeView._localRevisionHistoryItem.uri !=
                         uri1)) {
                        dialog_alert("Cannot perform diff between client file and a depot file on a different branch");
                        return;
                    }
                    uri1 = filepath + "#" + version1
                }
                break;
            case 'cvs':
                options = this._prefsSvc.getStringPref('cvsDiffOptions');
                if (diffToLocal) {
                    options += " -r " + version1
                }
                break;
            case 'svn':
                options = this._prefsSvc.getStringPref('svnDiffOptions');
                if (diffToLocal) {
                    options += " --revision " + version1
                }
                break;
            default:
                dialog_alert("Invalid Source Code Control Handler: "+scchandler);
                window.setCursor("auto");
                return;
        }
        if (diffToLocal) {
            var diff = this.sccHandler.diff(1, [ uri1 ], options, external);
        } else {
            var diff = this.sccHandler.diffRevisions(uri1, version1,
                                                     uri2, version2,
                                                     filepath, options,
                                                     external);
        }
        var error = this._lastErrorSvc.getLastErrorMessage();
        //output = this.generateCommandInfo(scchandler+' diff',urls,diff);
        //SCCScinAppendText(output);
        if (!error && external == '') {
            if (diff) {
                var diffDisplayStyle = this._prefsSvc.getStringPref('diffDisplayStyle');
                if (diffDisplayStyle == 'tab') {
                    try {
                        diffView = gViewMgr.doNewView('Diff');
                        diffView.document.buffer = diff;
                    } catch (e) {
                        log.exception(e);
                    }
                } else if (diffDisplayStyle == 'diffwin') {
                    // Put together a title for the diff output win.
                    var diff_title = this.sccType + " diff:";
                    diff_title += " "+uriparse_baseName(filepath);
                    launch_DiffWindow(diff, diff_title);
                }
            } else {
                alert("Files are in sync with repository.");
            }
        } else {
            dialog_alert("Error: " + error);
        }
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

_scc_history_dialog.prototype.diffToLocalVersion = function ()
{
    try {
        window.setCursor("wait");
        // Get the uri
        var filepath = this.urls[this.urlsTree.currentIndex];
        // Get the version number of the selected row
        var rowNum = this.historyTreeView.selection.currentIndex;
        var uri1 = this.historyTreeView._rows[rowNum].uri;
        var version1 = this.historyTreeView._rows[rowNum].version;

        this.performDiff(uri1, version1, '', '', filepath, true);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
    window.setCursor("auto");
}

_scc_history_dialog.prototype.diffToCurrentVersion = function ()
{
    try {
        window.setCursor("wait");
        // Get the uri
        var uri = this.urls[this.urlsTree.currentIndex];
        // Get the version number of the selected row
        var rowNum = this.historyTreeView.selection.currentIndex;
        var uri1 = this.historyTreeView._rows[rowNum].uri;
        var version1 = this.historyTreeView._rows[rowNum].version;

        this.performDiff(uri1, version1, filepath, '', filepath, false);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
    window.setCursor("auto");
}

_scc_history_dialog.prototype.diffBetweenVersions = function ()
{
    try {
        // Get the uri
        window.setCursor("wait");

        // Get the selected tree rows
        var selected = [];
        var start = new Object();
        var end = new Object();
        var numRanges = this.historyTreeView.selection.getRangeCount();
        for (var t=0; t<numRanges; t++) {
            this.historyTreeView.selection.getRangeAt(t, start, end);
            for (var v = start.value; v <= end.value; v++){
                selected.push(v);
            }
        }
        //alert("Selected: " + selected);
        // Get the version numbers of the selected rows
        // XXX - For perforce, we need to take into account branches, as this
        //       means they may be a different filenames.
        var uri1 = this.historyTreeView._rows[selected[1]].uri;
        var version1 = this.historyTreeView._rows[selected[1]].version;
        var uri2 = this.historyTreeView._rows[selected[0]].uri;
        var version2 = this.historyTreeView._rows[selected[0]].version;
        var filepath = this.urls[this.urlsTree.currentIndex];

        this.performDiff(uri1, version1, uri2, version2, filepath, false);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
    window.setCursor("auto");
}

_scc_history_dialog.prototype.doDefaultDiff = function ()
{
    try {
        if (this.historyTreeView.selection.count == 1) {
            this.diffToLocalVersion();
        } else if (this.historyTreeView.selection.count == 2) {
            this.diffBetweenVersions();
        }
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

_scc_history_dialog.prototype.handleKeypress = function (event)
{
    try {
        if (event.keyCode == event.DOM_VK_ENTER ||
            event.keyCode == event.DOM_VK_RETURN) {
            // We only want to fire this when we are not already focused on a
            // button, else it gets fired automatically by the button handler
            if (event.target.id != 'scc_history_diff_to_local_version_button' &&
                event.target.id != 'scc_history_diff_between_versions_button') {
                event.stopPropagation();
                event.preventDefault();
                event.cancelBubble = true;
                gSCCHistoryDialog.doDefaultDiff();
            }
        }
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

